# Copyright (c) OpenMMLab. All rights reserved.
import functools
import pickle
import warnings
from collections import OrderedDict

import numpy as np
import torch
import torch.distributed as dist
from antgo.framework.helper.runner import OptimizerHook, get_dist_info
from torch._utils import (_flatten_dense_tensors, _take_tensors,
                          _unflatten_dense_tensors)


def _allreduce_coalesced(tensors, world_size, bucket_size_mb=-1):
    if bucket_size_mb > 0:
        bucket_size_bytes = bucket_size_mb * 1024 * 1024
        buckets = _take_tensors(tensors, bucket_size_bytes)
    else:
        buckets = OrderedDict()
        for tensor in tensors:
            tp = tensor.type()
            if tp not in buckets:
                buckets[tp] = []
            buckets[tp].append(tensor)
        buckets = buckets.values()

    for bucket in buckets:
        flat_tensors = _flatten_dense_tensors(bucket)
        dist.all_reduce(flat_tensors)
        flat_tensors.div_(world_size)
        for tensor, synced in zip(
                bucket, _unflatten_dense_tensors(flat_tensors, bucket)):
            tensor.copy_(synced)


def allreduce_grads(params, coalesce=True, bucket_size_mb=-1):
    """Allreduce gradients.

    Args:
        params (list[torch.Parameters]): List of parameters of a model
        coalesce (bool, optional): Whether allreduce parameters as a whole.
            Defaults to True.
        bucket_size_mb (int, optional): Size of bucket, the unit is MB.
            Defaults to -1.
    """
    grads = [
        param.grad.data for param in params
        if param.requires_grad and param.grad is not None
    ]
    world_size = dist.get_world_size()
    if coalesce:
        _allreduce_coalesced(grads, world_size, bucket_size_mb)
    else:
        for tensor in grads:
            dist.all_reduce(tensor.div_(world_size))


class DistOptimizerHook(OptimizerHook):
    """Deprecated optimizer hook for distributed training."""

    def __init__(self, *args, **kwargs):
        warnings.warn('"DistOptimizerHook" is deprecated, please switch to'
                      '"mmcv.runner.OptimizerHook".')
        super().__init__(*args, **kwargs)


def reduce_mean(tensor):
    """"Obtain the mean of tensor on different GPUs."""
    if not (dist.is_available() and dist.is_initialized()):
        return tensor
    tensor = tensor.clone()
    dist.all_reduce(tensor.div_(dist.get_world_size()), op=dist.ReduceOp.SUM)
    return tensor


def obj2tensor(pyobj, device='cuda'):
    """Serialize picklable python object to tensor."""
    storage = torch.ByteStorage.from_buffer(pickle.dumps(pyobj))
    return torch.ByteTensor(storage).to(device=device)


def tensor2obj(tensor):
    """Deserialize tensor to picklable python object."""
    return pickle.loads(tensor.cpu().numpy().tobytes())


@functools.lru_cache()
def _get_global_gloo_group():
    """Return a process group based on gloo backend, containing all the ranks
    The result is cached."""
    if dist.get_backend() == 'nccl':
        return dist.new_group(backend='gloo')
    else:
        return dist.group.WORLD


def all_reduce_dict(py_dict, op='sum', group=None, to_float=True):
    """Apply all reduce function for python dict object.

    The code is modified from https://github.com/Megvii-
    BaseDetection/YOLOX/blob/main/yolox/utils/allreduce_norm.py.

    NOTE: make sure that py_dict in different ranks has the same keys and
    the values should be in the same shape. Currently only supports
    nccl backend.

    Args:
        py_dict (dict): Dict to be applied all reduce op.
        op (str): Operator, could be 'sum' or 'mean'. Default: 'sum'
        group (:obj:`torch.distributed.group`, optional): Distributed group,
            Default: None.
        to_float (bool): Whether to convert all values of dict to float.
            Default: True.

    Returns:
        OrderedDict: reduced python dict object.
    """
    warnings.warn(
        'group` is deprecated. Currently only supports NCCL backend.')
    _, world_size = get_dist_info()
    if world_size == 1:
        return py_dict

    # all reduce logic across different devices.
    py_key = list(py_dict.keys())
    if not isinstance(py_dict, OrderedDict):
        py_key_tensor = obj2tensor(py_key)
        dist.broadcast(py_key_tensor, src=0)
        py_key = tensor2obj(py_key_tensor)

    tensor_shapes = [py_dict[k].shape for k in py_key]
    tensor_numels = [py_dict[k].numel() for k in py_key]

    if to_float:
        warnings.warn('Note: the "to_float" is True, you need to '
                      'ensure that the behavior is reasonable.')
        flatten_tensor = torch.cat(
            [py_dict[k].flatten().float() for k in py_key])
    else:
        flatten_tensor = torch.cat([py_dict[k].flatten() for k in py_key])

    dist.all_reduce(flatten_tensor, op=dist.ReduceOp.SUM)
    if op == 'mean':
        flatten_tensor /= world_size

    split_tensors = [
        x.reshape(shape) for x, shape in zip(
            torch.split(flatten_tensor, tensor_numels), tensor_shapes)
    ]
    out_dict = {k: v for k, v in zip(py_key, split_tensors)}
    if isinstance(py_dict, OrderedDict):
        out_dict = OrderedDict(out_dict)
    return out_dict


def sync_random_seed(seed=None, device='cuda'):
    """Make sure different ranks share the same seed.

    All workers must call this function, otherwise it will deadlock.
    This method is generally used in `DistributedSampler`,
    because the seed should be identical across all processes
    in the distributed group.

    In distributed sampling, different ranks should sample non-overlapped
    data in the dataset. Therefore, this function is used to make sure that
    each rank shuffles the data indices in the same order based
    on the same seed. Then different ranks could use different indices
    to select non-overlapped data from the same data list.

    Args:
        seed (int, Optional): The seed. Default to None.
        device (str): The device where the seed will be put on.
            Default to 'cuda'.

    Returns:
        int: Seed to be used.
    """
    if seed is None:
        seed = np.random.randint(2**31)
    assert isinstance(seed, int)

    rank, world_size = get_dist_info()

    if world_size == 1:
        return seed

    if rank == 0:
        random_num = torch.tensor(seed, dtype=torch.int32, device=device)
    else:
        random_num = torch.tensor(0, dtype=torch.int32, device=device)
    dist.broadcast(random_num, src=0)
    return random_num.item()



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from ..builder import BBOX_SAMPLERS
from .base_sampler import BaseSampler


@BBOX_SAMPLERS.register_module()
class RandomSampler(BaseSampler):
    """Random sampler.

    Args:
        num (int): Number of samples
        pos_fraction (float): Fraction of positive samples
        neg_pos_up (int, optional): Upper bound number of negative and
            positive samples. Defaults to -1.
        add_gt_as_proposals (bool, optional): Whether to add ground truth
            boxes as proposals. Defaults to True.
    """

    def __init__(self,
                 num,
                 pos_fraction,
                 neg_pos_ub=-1,
                 add_gt_as_proposals=True,
                 **kwargs):
        from mmdet.core.bbox import demodata
        super(RandomSampler, self).__init__(num, pos_fraction, neg_pos_ub,
                                            add_gt_as_proposals)
        self.rng = demodata.ensure_rng(kwargs.get('rng', None))

    def random_choice(self, gallery, num):
        """Random select some elements from the gallery.

        If `gallery` is a Tensor, the returned indices will be a Tensor;
        If `gallery` is a ndarray or list, the returned indices will be a
        ndarray.

        Args:
            gallery (Tensor | ndarray | list): indices pool.
            num (int): expected sample num.

        Returns:
            Tensor or ndarray: sampled indices.
        """
        assert len(gallery) >= num

        is_tensor = isinstance(gallery, torch.Tensor)
        if not is_tensor:
            if torch.cuda.is_available():
                device = torch.cuda.current_device()
            else:
                device = 'cpu'
            gallery = torch.tensor(gallery, dtype=torch.long, device=device)
        # This is a temporary fix. We can revert the following code
        # when PyTorch fixes the abnormal return of torch.randperm.
        # See: https://github.com/open-mmlab/mmdetection/pull/5014
        perm = torch.randperm(gallery.numel())[:num].to(device=gallery.device)
        rand_inds = gallery[perm]
        if not is_tensor:
            rand_inds = rand_inds.cpu().numpy()
        return rand_inds

    def _sample_pos(self, assign_result, num_expected, **kwargs):
        """Randomly sample some positive samples."""
        pos_inds = torch.nonzero(assign_result.gt_inds > 0, as_tuple=False)
        if pos_inds.numel() != 0:
            pos_inds = pos_inds.squeeze(1)
        if pos_inds.numel() <= num_expected:
            return pos_inds
        else:
            return self.random_choice(pos_inds, num_expected)

    def _sample_neg(self, assign_result, num_expected, **kwargs):
        """Randomly sample some negative samples."""
        neg_inds = torch.nonzero(assign_result.gt_inds == 0, as_tuple=False)
        if neg_inds.numel() != 0:
            neg_inds = neg_inds.squeeze(1)
        if len(neg_inds) <= num_expected:
            return neg_inds
        else:
            return self.random_choice(neg_inds, num_expected)



# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn as nn
from .conv_module import ConvModule
from .make_divisible import make_divisible


class SELayer(nn.Module):
    """Squeeze-and-Excitation Module.

    Args:
        channels (int): The input (and output) channels of the SE layer.
        squeeze_channels (None or int): The intermediate channel number of
            SElayer. Default: None, means the value of ``squeeze_channels``
            is ``make_divisible(channels // ratio, divisor)``.
        ratio (int): Squeeze ratio in SELayer, the intermediate channel will
            be ``make_divisible(channels // ratio, divisor)``. Only used when
            ``squeeze_channels`` is None. Default: 16.
        divisor(int): The divisor to true divide the channel number. Only
            used when ``squeeze_channels`` is None. Default: 8.
        conv_cfg (None or dict): Config dict for convolution layer. Default:
            None, which means using conv2d.
        return_weight(bool): Whether to return the weight. Default: False.
        act_cfg (dict or Sequence[dict]): Config dict for activation layer.
            If act_cfg is a dict, two activation layers will be configurated
            by this dict. If act_cfg is a sequence of dicts, the first
            activation layer will be configurated by the first dict and the
            second activation layer will be configurated by the second dict.
            Default: (dict(type='ReLU'), dict(type='Sigmoid'))
    """

    def __init__(self,
                 channels,
                 squeeze_channels=None,
                 ratio=16,
                 divisor=8,
                 bias='auto',
                 conv_cfg=None,
                 act_cfg=(dict(type='ReLU'), dict(type='Sigmoid')),
                 return_weight=False):
        super(SELayer, self).__init__()
        if isinstance(act_cfg, dict):
            act_cfg = (act_cfg, act_cfg)
        assert len(act_cfg) == 2
        self.global_avgpool = nn.AdaptiveAvgPool2d(1)
        if squeeze_channels is None:
            squeeze_channels = make_divisible(channels // ratio, divisor)
        assert isinstance(squeeze_channels, int) and squeeze_channels > 0, \
            '"squeeze_channels" should be a positive integer, but get ' + \
            f'{squeeze_channels} instead.'
        self.return_weight = return_weight
        self.conv1 = ConvModule(
            in_channels=channels,
            out_channels=squeeze_channels,
            kernel_size=1,
            stride=1,
            bias=bias,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg[0])
        self.conv2 = ConvModule(
            in_channels=squeeze_channels,
            out_channels=channels,
            kernel_size=1,
            stride=1,
            bias=bias,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg[1])

    def forward(self, x):
        out = self.global_avgpool(x)
        out = self.conv1(out)
        out = self.conv2(out)
        if self.return_weight:
            return out
        else:
            return x * out



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/11 23:10
# @File    : dataframe.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import json

from typing import Dict, Any, Optional, Set, Union, List

from antgo.pipeline.functional.entity import Entity
from antgo.pipeline.hparam import dynamic_dispatch, param_scope


class DataFrameMixin:
    """
    Mixin to help deal with Entity.

    Examples:

    1. define an operators with `register` decorator

    >>> @register
    ... def add_1(x):
    ...     return x+1

    2. apply the operators to named field of entity and save result to another named field

    >>> (
    ...     DataFrame([dict(a=1, b=2), dict(a=2, b=3)])
    ...         .as_entity()
    ...         .add_1['a', 'c']() # <-- use field `a` as input and filed `c` as output
    ...         .as_str()
    ...         .to_list()
    ... )
    ["{'a': 1, 'b': 2, 'c': 2}", "{'a': 2, 'b': 3, 'c': 3}"]

    Select the entity on the specified fields.

    Examples:

    1. Select the entity on one specified field:

    >>> df = DataFrame([Entity(a=i, b=i, c=i) for i in range(2)])
    >>> df.select['a']().to_list()
    [<Entity dict_keys(['a'])>, <Entity dict_keys(['a'])>]

    2. Select multiple fields and unpack the entity:

    >>> (
    ...     DataFrame([Entity(a=i, b=i, c=i) for i in range(5)])
    ...         .select['a', 'b']()
    ...         .as_raw()
    ...         .to_list()
    ... )
    [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4)]

    3. Another field selection syntax (not suggested):

    >>> (
    ...     DataFrame([Entity(a=i, b=i, c=i) for i in range(5)])
    ...         .select('a', 'b')
    ...         .as_raw()
    ...         .to_list()
    ... )
    [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4)]
    """

    def __init__(self):
        # pylint: disable=useless-super-delegation
        super().__init__()

    @property
    def select(self):

        @dynamic_dispatch
        def selector(*arg):
            index = param_scope()._index
            if isinstance(index, str):
                index = (index, )
            if index is None and arg is not None and len(arg) > 0:
                index = arg

            def inner(entity: Entity):
                if index is not None:
                    return Entity(
                        **{col: getattr(entity, col)
                           for col in index})

                return entity

            return self.map(inner)

        return selector

    # pylint: disable=invalid-name
    def fill_entity(self,
                    _DefaultKVs: Optional[Dict[str, Any]] = None,
                    _ReplaceNoneValue: bool = False,
                    **kws):
        """
        When DataFrame's iterable exists of Entities and some indexes missing, fill default value for those indexes.

        Args:
            _ReplaceNoneValue (`bool`):
                Whether to replace None in Entity's value.
            _DefaultKVs (`Dict[str, Any]`):
                The key-value pairs stored in a dict.

        Examples:

        >>> entities = [Entity(num=i) for i in range(3)]
        >>> df = DataFrame(entities)
        >>> df
        [<Entity dict_keys(['num'])>, <Entity dict_keys(['num'])>, <Entity dict_keys(['num'])>]

        >>> kvs = {'foo': 'bar'}
        >>> df.fill_entity(kvs).fill_entity(usage='test').to_list()
        [<Entity dict_keys(['num', 'foo', 'usage'])>, <Entity dict_keys(['num', 'foo', 'usage'])>, <Entity dict_keys(['num', 'foo', 'usage'])>]

        >>> kvs = {'FOO': None}
        >>> df.fill_entity(_ReplaceNoneValue=True, _DefaultKVs=kvs).to_list()[0].FOO
        0
        """
        if _DefaultKVs:
            kws.update(_DefaultKVs)

        def fill(entity: Entity):
            for k, v in kws.items():
                if not hasattr(entity, k):
                    setattr(entity, k, v)
                if _ReplaceNoneValue and v is None:
                    setattr(entity, k, 0)
            return entity

        return self._factory(map(fill, self._iterable))

    def as_entity(self, schema: Optional[List[str]] = None):
        """
        Convert elements into Entities.

        Args:
            schema (Optional[List[str]]):
                schema contains field names.

        Examples:
        1. convert dicts into entities:

        >>> (
        ...     DataFrame([dict(a=1, b=2), dict(a=2, b=3)])
        ...         .as_entity()
        ...         .as_str()
        ...         .to_list()
        ... )
        ["{'a': 1, 'b': 2}", "{'a': 2, 'b': 3}"]

        2. convert tuples into entities:

        >>> (
        ...     DataFrame([(1, 2), (2, 3)])
        ...         .as_entity(schema=['a', 'b'])
        ...         .as_str()
        ...         .to_list()
        ... )
        ["{'a': 1, 'b': 2}", "{'a': 2, 'b': 3}"]

        3. convert single value into entities:

        >>> (
        ...     DataFrame([1, 2])
        ...         .as_entity(schema=['a'])
        ...         .as_str()
        ...         .to_list()
        ... )
        ["{'a': 1}", "{'a': 2}"]
        """

        if schema is None:

            def inner(x):
                return Entity(**x)
        else:

            def inner(x):
                if len(schema) == 1:
                    x = (x, )
                data = dict(zip(schema, x))
                return Entity(**data)

        return self._factory(map(inner, self._iterable))

    def parse_json(self):
        """
        Parse string to entities.

        Examples:

        >>> df = (
        ...     DataFrame(['{"x": 1}'])
        ...         .parse_json()
        ... )
        >>> df[0].x
        1
        """

        def inner(x):
            data = json.loads(x)
            return Entity(**data)

        return self.map(inner)

    def as_json(self):
        """
        Convert entities to json

        Examples:

        >>> (
        ...     DataFrame([Entity(x=1)])
        ...         .as_json()
        ... )
        ['{"x": 1}']
        """

        def inner(x):
            return json.dumps(x.__dict__)

        return self.map(inner)

    def as_raw(self):
        """
        Convert entitis into raw python values

        Examples:

        1. unpack multiple values from entities:

        >>> (
        ...     DataFrame([(1, 2), (2, 3)])
        ...         .as_entity(schema=['a', 'b'])
        ...         .as_raw()
        ...         .to_list()
        ... )
        [(1, 2), (2, 3)]

        2. unpack single value from entities:

        >>> (
        ...     DataFrame([1, 2])
        ...         .as_entity(schema=['a'])
        ...         .as_raw()
        ...         .to_list()
        ... )
        [1, 2]
        """

        def inner(x):
            if len(x.__dict__) == 1:
                return list(x.__dict__.values())[0]
            return tuple(getattr(x, name) for name in x.__dict__)

        return self.map(inner)

    def replace(self, **kws):
        """
        Replace specific attributes with given vlues.

        Examples:


        >>> entities = [Entity(num=i) for i in range(5)]
        >>> df = DataFrame(entities)
        >>> [i.num for i in df]
        [0, 1, 2, 3, 4]

        >>> df = df.replace(num={0: 1, 1: 2, 2: 3, 3: 4, 4: 5})
        >>> [i.num for i in df]
        [1, 2, 3, 4, 5]
        """

        def inner(entity: Entity):
            for index, convert_dict in kws.items():
                origin_value = getattr(entity, index)
                if origin_value in convert_dict:
                    setattr(entity, index, convert_dict[origin_value])

            return entity

        return self._factory(map(inner, self._iterable))

    def dropna(self, na: Set[str] = {'', None}) -> Union[bool, 'DataFrame']:  # pylint: disable=dangerous-default-value
        """
        Drop entities that contain some specific values.

        Args:
            na (`Set[str]`):
                Those entities contain values in na will be dropped.

        Examples:

        >>> entities = [Entity(a=i, b=i + 1) for i in range(3)]
        >>> entities.append(Entity(a=3, b=''))
        >>> df = DataFrame(entities)
        >>> df
        [<Entity dict_keys(['a', 'b'])>, <Entity dict_keys(['a', 'b'])>, <Entity dict_keys(['a', 'b'])>, <Entity dict_keys(['a', 'b'])>]

        >>> df.dropna()
        [<Entity dict_keys(['a', 'b'])>, <Entity dict_keys(['a', 'b'])>, <Entity dict_keys(['a', 'b'])>]
        """

        def inner(entity: Entity):
            for val in entity.__dict__.values():
                if val in na:
                    return False

            return True

        return self._factory(filter(inner, self._iterable))

    def rename(self, column: Dict[str, str]):
        """
        Rename an column in DataFrame.

        Args:
            column (`Dict[str, str]`):
                The columns to rename and their corresponding new name.

        Examples:

        >>> entities = [Entity(a=i, b=i + 1) for i in range(3)]
        >>> df = DataFrame(entities)
        >>> df
        [<Entity dict_keys(['a', 'b'])>, <Entity dict_keys(['a', 'b'])>, <Entity dict_keys(['a', 'b'])>]

        >>> df.rename(column={'a': 'A', 'b': 'B'})
        [<Entity dict_keys(['A', 'B'])>, <Entity dict_keys(['A', 'B'])>, <Entity dict_keys(['A', 'B'])>]
        """

        def inner(x):
            for key in column:
                x.__dict__[column[key]] = x.__dict__.pop(key)
            return x

        return self._factory(map(inner, self._iterable))

    @property
    def df(self):
        # pylint: disable=import-outside-toplevel
        import pandas as pd
        if isinstance(self._iterable, pd.DataFrame):
            return self._iterable
        else:
            raise TypeError(
                'data collection is not created from pandas DataFrame.')



import logging

import torch.distributed as dist

logger_initialized = {}


def get_logger(name, log_file=None, log_level=logging.INFO, file_mode='w'):
    """Initialize and get a logger by name.

    If the logger has not been initialized, this method will initialize the
    logger by adding one or two handlers, otherwise the initialized logger will
    be directly returned. During initialization, a StreamHandler will always be
    added. If `log_file` is specified and the process rank is 0, a FileHandler
    will also be added.

    Args:
        name (str): Logger name.
        log_file (str | None): The log filename. If specified, a FileHandler
            will be added to the logger.
        log_level (int): The logger level. Note that only the process of
            rank 0 is affected, and other processes will set the level to
            "Error" thus be silent most of the time.
        file_mode (str): The file mode used in opening log file.
            Defaults to 'w'.

    Returns:
        logging.Logger: The expected logger.
    """
    logger = logging.getLogger(name)
    if name in logger_initialized:
        return logger
    # handle hierarchical names
    # e.g., logger "a" is initialized, then logger "a.b" will skip the
    # initialization since it is a child of "a".
    for logger_name in logger_initialized:
        if name.startswith(logger_name):
            return logger

    # handle duplicate logs to the console
    # Starting in 1.8.0, PyTorch DDP attaches a StreamHandler <stderr> (NOTSET)
    # to the root logger. As logger.propagate is True by default, this root
    # level handler causes logging messages from rank>0 processes to
    # unexpectedly show up on the console, creating much unwanted clutter.
    # To fix this issue, we set the root logger's StreamHandler, if any, to log
    # at the ERROR level.
    for handler in logger.root.handlers:
        if type(handler) is logging.StreamHandler:
            handler.setLevel(logging.ERROR)

    stream_handler = logging.StreamHandler()
    handlers = [stream_handler]

    if dist.is_available() and dist.is_initialized():
        rank = dist.get_rank()
    else:
        rank = 0

    # only rank 0 will add a FileHandler
    if rank == 0 and log_file is not None:
        # Here, the default behaviour of the official logger is 'a'. Thus, we
        # provide an interface to change the file mode to the default
        # behaviour.
        file_handler = logging.FileHandler(log_file, file_mode)
        handlers.append(file_handler)

    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    for handler in handlers:
        handler.setFormatter(formatter)
        handler.setLevel(log_level)
        logger.addHandler(handler)

    if rank == 0:
        logger.setLevel(log_level)
    else:
        logger.setLevel(logging.ERROR)

    logger_initialized[name] = True

    return logger


def print_log(msg, logger=None, level=logging.INFO):
    """Print a log message.

    Args:
        msg (str): The message to be logged.
        logger (logging.Logger | str | None): The logger to be used.
            Some special loggers are:

            - "silent": no message will be printed.
            - other str: the logger obtained with `get_root_logger(logger)`.
            - None: The `print()` method will be used to print log messages.
        level (int): Logging level. Only available when `logger` is a Logger
            object or "root".
    """
    if logger is None:
        print(msg)
    elif isinstance(logger, logging.Logger):
        logger.log(level, msg)
    elif logger == 'silent':
        pass
    elif isinstance(logger, str):
        _logger = get_logger(logger)
        _logger.log(level, msg)
    else:
        raise TypeError(
            'logger should be either a logging.Logger object, str, '
            f'"silent" or None, but got {type(logger)}')



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES


'''
Adopted from https://github.com/passalis/probabilistic_kt/blob/master/nn/pkt.py
'''

@DISTILL_LOSSES.register_module()
class PKTCosSim(nn.Module):
	'''
	Learning Deep Representations with Probabilistic Knowledge Transfer
	http://openaccess.thecvf.com/content_ECCV_2018/papers/Nikolaos_Passalis_Learning_Deep_Representations_ECCV_2018_paper.pdf
	'''
	def __init__(self):
		super(PKTCosSim, self).__init__()

	def forward(self, feat_s, feat_t, eps=1e-6):
		# Normalize each vector by its norm
		feat_s_norm = torch.sqrt(torch.sum(feat_s ** 2, dim=1, keepdim=True))
		feat_s = feat_s / (feat_s_norm + eps)
		feat_s[feat_s != feat_s] = 0

		feat_t_norm = torch.sqrt(torch.sum(feat_t ** 2, dim=1, keepdim=True))
		feat_t = feat_t / (feat_t_norm + eps)
		feat_t[feat_t != feat_t] = 0

		# Calculate the cosine similarity
		feat_s_cos_sim = torch.mm(feat_s, feat_s.transpose(0, 1))
		feat_t_cos_sim = torch.mm(feat_t, feat_t.transpose(0, 1))

		# Scale cosine similarity to [0,1]
		feat_s_cos_sim = (feat_s_cos_sim + 1.0) / 2.0
		feat_t_cos_sim = (feat_t_cos_sim + 1.0) / 2.0

		# Transform them into probabilities
		feat_s_cond_prob = feat_s_cos_sim / torch.sum(feat_s_cos_sim, dim=1, keepdim=True)
		feat_t_cond_prob = feat_t_cos_sim / torch.sum(feat_t_cos_sim, dim=1, keepdim=True)

		# Calculate the KL-divergence
		loss = torch.mean(feat_t_cond_prob * torch.log((feat_t_cond_prob + eps) / (feat_s_cond_prob + eps)))

		return loss


import sys
sys.path.append('/root/workspace/antgo')
from antgo.dataflow.datasetio import *
import os
import json
import pickle
from antgo.framework.helper.dataset.tfdataset import *
from antgo.framework.helper.dataset.kvdataset import *
from antgo.ant import environment
import requests
import logging
from queue import Queue
import threading
import time
import copy


class __SampleDataGenerator(object):
    def __init__(self, json_files, thread_num=1, mode='json') -> None:
        self.json_file_list = json_files.split(',')
        self.mode = mode    # line or json. line: 意味着每一行是一个json格式，json: 意味着整体是一个json格式

        # 预先加载，获得样本总数
        self.num_sample = 0
        self.num_sample_offset = [0]
        self.num_sample_list = []
        for json_file in self.json_file_list:
            if self.mode == 'json':
                with open(json_file, 'r') as fp:
                    content = json.load(fp)
                    self.num_sample += len(content)
                    self.num_sample_offset.append(self.num_sample_offset[-1]+len(content))
                    self.num_sample_list.append(len(content))
            else:
                with open(json_file, 'r') as fp:
                    content = fp.readline()
                    content = content.strip()

                    num = 0
                    while content:
                        num += 1
                        content = fp.readline()
                        content = content.strip()
                        if content == '':
                            break

                    self.num_sample += num
                    self.num_sample_offset.append(self.num_sample_offset[-1]+num)
                    self.num_sample_list.append(num)

        print(f"Sample num {self.num_sample}")
        
        # 每个线程 分配的样本数
        assign_sample_num = (self.num_sample + (thread_num - 1)) // thread_num

        # 线程标记
        self.thread_stop_flag = [1 for _ in range(thread_num)]
        # 启动线程池进行数据生产
        # 数据队列 (队列最大容量10000)
        self.data_cache = Queue(100)        
        self.thread_pool = [
            threading.Thread(target=self.produce, args=(i*assign_sample_num, min((i+1)*assign_sample_num, self.num_sample), i, self.data_cache)) for i in range(thread_num)]

        # 启动线程
        for thread_i in range(thread_num):
            self.thread_pool[thread_i].start()

    def __process_sample(self, src_folder, sample, data_cache):
        # 转换通用数据格式
        for k in sample.keys():
            if k in ['image_url', 'image_file']:
                continue

            # 对于list数据直接转换为numpy
            if isinstance(sample[k], list):
                if len(sample[k]) > 0:
                    # 如果存在二级list，则需要兼容空情况
                    if isinstance(sample[k][0], list):
                        max_num = 0
                        for i in range(len(sample[k])):
                            if max_num < len(sample[k][i]):
                                max_num = len(sample[k][i])
                        
                        for i in range(len(sample[k])):
                            if len(sample[k][i]) == 0:
                                sample[k][i] = [0 for _ in range(max_num)]
                else:
                    # 空list处理方法
                    sample[k] = np.array([], dtype=np.float32)

                # 转换到numpy类型    
                sample[k] = np.array(sample[k])
                # 转换浮点数值
                if sample[k].dtype in [np.float64, np.float16]:
                    sample[k] = sample[k].astype(np.float32)
                # 转换bool数值
                if sample[k].dtype in [np.bool8]:
                    sample[k] = sample[k].astype(np.int8)

            # 对bool数据转换成int
            if isinstance(sample[k], bool):
                sample[k] = int(sample[k])

        # 加载图像
        if 'image_url' not in sample and 'image_file' not in sample:
            logging.error('dont have image_url and image_file')
            return
        
        if sample['image_url'] == '' and sample['image_file'] == '':
            logging.error('dont have image_url and image_file')
            return
        
        if sample['image_url'] != '':
            wait_count = 5
            while wait_count > 0:
                try:
                    pic = requests.get(sample['image_url'], timeout=20)
                    if pic.status_code != 200:
                        sample['image'] = b''
                        wait_count -= 1
                        time.sleep(2)
                        continue

                    sample['image'] = pic.content
                    break
                except:
                    logging.error("Couldnt download %s."%sample['image_url'])
                    sample['image'] = b''
                    wait_count -= 1
                    time.sleep(2)               
        else:
            try:
                image_path = os.path.join(src_folder, sample['image_file'])
                if os.path.exists(image_path):
                    with open(image_path, 'rb') as fp:
                        image_content = fp.read()
                        sample['image'] = image_content
                else:
                    logging.error(f'Missing image {image_path}')
                    sample['image'] = b''
            except:
                logging.error("Failt load image_path")
                sample['image'] = b''

        if sample['image'] == b'':
            # 图像为空时，直接忽略样本
            logging.error(f"Sample {sample['image_file']} data abnormal")
            return
        else:
            # yield sample
            data_cache.put(sample)

    def line_mode_produce(self, start_ii, stop_ii, json_file, data_cache, src_folder):
        with open(json_file, 'r') as fp:
            content = fp.readline()
            content = content.strip()

            sample_i = 0
            while content:
                if sample_i < start_ii:
                    # 跳过样本
                    content = fp.readline()
                    content = content.strip()
                    sample_i += 1
                    if content == '':
                        break        
                    continue
                if sample_i >= stop_ii:
                    # 退出
                    break

                # 格式化一条样本
                sample = json.loads(content)
                
                # 读取下一条样本数据
                content = fp.readline()
                content = content.strip()
                sample_i += 1
                if content == '':
                    break

                self.__process_sample(src_folder, sample, data_cache)
    
    def json_mode_produce(self, start_ii, stop_ii, json_file, data_cache, src_folder):
        with open(json_file, 'r', encoding="utf-8") as fp:
            samples = json.load(fp)[start_ii:stop_ii]

        num = len(samples)
        for sample_i in range(num):
            sample = copy.deepcopy(samples[sample_i])

            self.__process_sample(src_folder, sample, data_cache)

    def produce(self, start_i, stop_i, thread_id, data_cache):
        # 发现从start_i 到 stop_i，来自于的json_file集合
        start_file_i = 0
        start_sample_offset = 0
        stop_file_i = 0
        stop_sample_offset = 0
        for file_i in range(len(self.json_file_list)):
            if start_i >= self.num_sample_offset[file_i] and start_i <= self.num_sample_offset[file_i+1]:
                start_file_i = file_i
                start_sample_offset = start_i - self.num_sample_offset[file_i]

            if stop_i >= self.num_sample_offset[file_i] and stop_i <= self.num_sample_offset[file_i+1]:
                stop_file_i = file_i
                stop_sample_offset = stop_i - self.num_sample_offset[file_i]

        for file_i in range(start_file_i, stop_file_i+1):
            json_file = self.json_file_list[file_i]
            src_folder = os.path.dirname(json_file)
            logging.info(f'Process {src_folder}')

            start_ii = 0
            stop_ii = self.num_sample_list[file_i]
            if file_i == start_file_i:
                start_ii = start_sample_offset

            if file_i == stop_file_i:
                stop_ii = min(stop_sample_offset, self.num_sample_list[file_i])

            print(f'thread_id {thread_id} start_ii {start_ii} stop_ii {stop_ii} in file_i {file_i}')
            if self.mode == 'line':
                self.line_mode_produce(start_ii, stop_ii, json_file, data_cache, src_folder)
            else:
                self.json_mode_produce(start_ii, stop_ii, json_file, data_cache, src_folder)

        # 设置线程标记
        self.thread_stop_flag[thread_id] = 0

    def __len__(self):
        return self.num_sample

    def __iter__(self):
        while np.sum(self.thread_stop_flag) > 0:
            try:
                data = self.data_cache.get(True, timeout=5)   # 最大等待5s
            except:
                continue

            if data is not None:
                yield data


def package_to_kv(src_file, tgt_folder, prefix, size_in_shard=-1, **kwargs):
    # src_file json 文件 (仅支持标准格式 sample_gt.json)
    if not os.path.exists(tgt_folder):
        os.makedirs(tgt_folder)

    # 创建writer 实例
    kvw = KVDataWriter(prefix, tgt_folder, -1)

    # 写数据
    kvw.write(__SampleDataGenerator(src_file, thread_num=kwargs.get('thread_num', 10), mode=kwargs.get('mode', 'json')))


def package_to_tfrecord(src_file, tgt_folder, prefix, size_in_shard=-1, **kwargs):
    # src_file json 文件 (仅支持标准格式 sample_gt.json)
    if not os.path.exists(tgt_folder):
        os.makedirs(tgt_folder)

    # 创建tfdatawriter 实例
    if size_in_shard <= 0:
        size_in_shard = 100000
    tfw = TFDataWriter(prefix, tgt_folder, size_in_shard)

    # 写数据
    tfw.write(__SampleDataGenerator(src_file, thread_num=kwargs.get('thread_num', 10), mode=kwargs.get('mode', 'json')))


# test tfrecord
# with open('/root/workspace/extract/ss.json', 'r') as fp:
#     content = json.load(fp)
#     # content = set(content)
#     print(len(content))

# start_time = time.time()
# package_to_tfrecord("/root/workspace/shuffle-hard-mining-v5/annotation_label_0.json,/root/workspace/shuffle-hard-mining-v5/annotation_label_0.json,/root/workspace/shuffle-hard-mining-v5/annotation_label_0.json", "/root/workspace/dataset/extract", "hello", 100000, thread_num=2, mode='json')
# print(f'all time {time.time() - start_time}')

# print('sdf')

# tfd = TFDataset(data_path_list=['/root/workspace/BB/hello-00000-of-00001-tfrecord'], shuffle_queue_size=2)
# for a in tfd:
#     print(a)

# test kv
# package_to_kv("/root/workspace/ss/annotation.json", "/root/workspace/BB", "hello")
# kvd = KVDatasetReader(data_path_list=['/root/workspace/BB/hello-00000-of-00001-kvrecord'])
# print(len(kvd))

# index = [10,20]
# result = kvd.reads(index)
# print(result)


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 20:52
# @File    : stream.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from typing import Iterator

from antgo.pipeline.functional.mixins.dag import register_dag


class StreamMixin:
    """
    Stream related mixins.
    """
    @register_dag
    def stream(self):
        """
        Create a stream data collection.

        Examples:
        1. Convert a data collection to streamed version

        >>> dc = DataCollection([0, 1, 2, 3, 4])
        >>> dc.is_stream
        False

        >>> dc = dc.stream()
        >>> dc.is_stream
        True
        """
        # pylint: disable=protected-access
        iterable = iter(self._iterable) if not self.is_stream else self._iterable
        return self._factory(iterable, parent_stream=False)

    @register_dag
    def unstream(self):
        """
        Create a unstream data collection.

        Examples:

        1. Create a unstream data collection

        >>> dc = DataCollection(iter(range(5))).unstream()
        >>> dc.is_stream
        False

        2. Convert a streamed data collection to unstream version

        >>> dc = DataCollection(iter(range(5)))
        >>> dc.is_stream
        True
        >>> dc = dc.unstream()
        >>> dc.is_stream
        False
        """
        iterable = list(self._iterable) if self.is_stream else self._iterable
        return self._factory(iterable, parent_stream=False)

    @property
    def is_stream(self):
        """
        Check whether the data collection is stream or unstream.

        Examples:

        >>> from typing import Iterable
        >>> dc = DataCollection([0,1,2,3,4])
        >>> dc.is_stream
        False

        >>> result = dc.map(lambda x: x+1)
        >>> result.is_stream
        False
        >>> result._iterable
        [1, 2, 3, 4, 5]

        >>> dc = DataCollection(iter(range(5)))
        >>> dc.is_stream
        True

        >>> result = dc.map(lambda x: x+1)
        >>> result.is_stream
        True
        >>> isinstance(result._iterable, Iterable)
        True

        """
        return isinstance(self._iterable, Iterator)



# 优化器
optimizer = dict(
    model=dict(),                   # 由具体下游任务定义
    discriminator=dict(
        type='SGD', lr=0.001, weight_decay=1e-4, momentum=0.9, nesterov=True
    )
)

# 学习率
lr_config = dict(
    model=dict(),                   # 由具体下游任务定义
    discriminator=dict(
        policy='CosineAnnealing',
        min_lr=1e-5,
    )
)

# 模型配置
model = dict(
    type='Adda',
    model=dict(),              # 由具体下游任务定义    
    train_cfg=dict(
        input_dim=1024,
        latent_dim=512,
        src_domain_batch_size=5,
        gen_loss_weight=1.0,
        dis_loss_weight=1.0,
        encoder_config=dict(
            from_name='backbone',
            from_index=-1
        )
    )
)


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/15 23:19
# @File    : show.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.interactcontext import *
from antgo.pipeline.hparam import param_scope


class ShowMixin:
  def browser(self, type={}, title='antgo'):
    ctx = InteractContext()
    with ctx.Browser(title, {}, browser={'size': 0}) as hp:
      for index, info in enumerate(self):
        data = {}

        for k, v in type.items():
          data[k] = {
            'data': getattr(info, k),
            'type': v
          }

        data['id'] = index
        hp.context.recorder.record(data)

      hp.wait_until_stop()

  def label(self, label_type='RECT', label_metas={}, title='antgo'):
    ctx = InteractContext()
    with ctx.Activelearning(title, {}, activelearning={
      'label_type': label_type,
      'label_metas': label_metas,
      'stage': 'labeling'
    }, clear=False) as activelearning:
      for index, info in enumerate(self):
        data = {
          'image': None,  # 第二优先级
          'label_info': [],
          'id': index
        }
        for k in info.__dict__.keys():
          data['image'] = getattr(info, k)
          break
        activelearning.context.recorder.record(data)

      activelearning.wait_until_stop()


import torch
import torch.nn as nn
import torch.nn.functional as F

from antgo.framework.helper.models.builder import LOSSES
from .accuracy import accuracy
from .cross_entropy_loss import cross_entropy
from .utils import weight_reduce_loss


def seesaw_ce_loss(cls_score,
                   labels,
                   label_weights,
                   cum_samples,
                   num_classes,
                   p,
                   q,
                   eps,
                   reduction='mean',
                   avg_factor=None):
    """Calculate the Seesaw CrossEntropy loss.

    Args:
        cls_score (torch.Tensor): The prediction with shape (N, C),
             C is the number of classes.
        labels (torch.Tensor): The learning label of the prediction.
        label_weights (torch.Tensor): Sample-wise loss weight.
        cum_samples (torch.Tensor): Cumulative samples for each category.
        num_classes (int): The number of classes.
        p (float): The ``p`` in the mitigation factor.
        q (float): The ``q`` in the compenstation factor.
        eps (float): The minimal value of divisor to smooth
             the computation of compensation factor
        reduction (str, optional): The method used to reduce the loss.
        avg_factor (int, optional): Average factor that is used to average
            the loss. Defaults to None.

    Returns:
        torch.Tensor: The calculated loss
    """
    assert cls_score.size(-1) == num_classes
    assert len(cum_samples) == num_classes

    onehot_labels = F.one_hot(labels, num_classes)
    seesaw_weights = cls_score.new_ones(onehot_labels.size())

    # mitigation factor
    if p > 0:
        sample_ratio_matrix = cum_samples[None, :].clamp(
            min=1) / cum_samples[:, None].clamp(min=1)
        index = (sample_ratio_matrix < 1.0).float()
        sample_weights = sample_ratio_matrix.pow(p) * index + (1 - index)
        mitigation_factor = sample_weights[labels.long(), :]
        seesaw_weights = seesaw_weights * mitigation_factor

    # compensation factor
    if q > 0:
        scores = F.softmax(cls_score.detach(), dim=1)
        self_scores = scores[
            torch.arange(0, len(scores)).to(scores.device).long(),
            labels.long()]
        score_matrix = scores / self_scores[:, None].clamp(min=eps)
        index = (score_matrix > 1.0).float()
        compensation_factor = score_matrix.pow(q) * index + (1 - index)
        seesaw_weights = seesaw_weights * compensation_factor

    cls_score = cls_score + (seesaw_weights.log() * (1 - onehot_labels))

    loss = F.cross_entropy(cls_score, labels, weight=None, reduction='none')

    if label_weights is not None:
        label_weights = label_weights.float()
    loss = weight_reduce_loss(
        loss, weight=label_weights, reduction=reduction, avg_factor=avg_factor)
    return loss


@LOSSES.register_module()
class SeesawLoss(nn.Module):
    """
    Seesaw Loss for Long-Tailed Instance Segmentation (CVPR 2021)
    arXiv: https://arxiv.org/abs/2008.10032

    Args:
        use_sigmoid (bool, optional): Whether the prediction uses sigmoid
             of softmax. Only False is supported.
        p (float, optional): The ``p`` in the mitigation factor.
             Defaults to 0.8.
        q (float, optional): The ``q`` in the compenstation factor.
             Defaults to 2.0.
        num_classes (int, optional): The number of classes.
             Default to 1203 for LVIS v1 dataset.
        eps (float, optional): The minimal value of divisor to smooth
             the computation of compensation factor
        reduction (str, optional): The method that reduces the loss to a
             scalar. Options are "none", "mean" and "sum".
        loss_weight (float, optional): The weight of the loss. Defaults to 1.0
        return_dict (bool, optional): Whether return the losses as a dict.
             Default to True.
    """

    def __init__(self,
                 use_sigmoid=False,
                 p=0.8,
                 q=2.0,
                 num_classes=1203,
                 eps=1e-2,
                 reduction='mean',
                 loss_weight=1.0,
                 return_dict=True):
        super(SeesawLoss, self).__init__()
        assert not use_sigmoid
        self.use_sigmoid = False
        self.p = p
        self.q = q
        self.num_classes = num_classes
        self.eps = eps
        self.reduction = reduction
        self.loss_weight = loss_weight
        self.return_dict = return_dict

        # 0 for pos, 1 for neg
        self.cls_criterion = seesaw_ce_loss

        # cumulative samples for each category
        self.register_buffer(
            'cum_samples',
            torch.zeros(self.num_classes + 1, dtype=torch.float))

        # custom output channels of the classifier
        self.custom_cls_channels = True
        # custom activation of cls_score
        self.custom_activation = True
        # custom accuracy of the classsifier
        self.custom_accuracy = True

    def _split_cls_score(self, cls_score):
        # split cls_score to cls_score_classes and cls_score_objectness
        assert cls_score.size(-1) == self.num_classes + 2
        cls_score_classes = cls_score[..., :-2]
        cls_score_objectness = cls_score[..., -2:]
        return cls_score_classes, cls_score_objectness

    def get_cls_channels(self, num_classes):
        """Get custom classification channels.

        Args:
            num_classes (int): The number of classes.

        Returns:
            int: The custom classification channels.
        """
        assert num_classes == self.num_classes
        return num_classes + 2

    def get_activation(self, cls_score):
        """Get custom activation of cls_score.

        Args:
            cls_score (torch.Tensor): The prediction with shape (N, C + 2).

        Returns:
            torch.Tensor: The custom activation of cls_score with shape
                 (N, C + 1).
        """
        cls_score_classes, cls_score_objectness = self._split_cls_score(
            cls_score)
        score_classes = F.softmax(cls_score_classes, dim=-1)
        score_objectness = F.softmax(cls_score_objectness, dim=-1)
        score_pos = score_objectness[..., [0]]
        score_neg = score_objectness[..., [1]]
        score_classes = score_classes * score_pos
        scores = torch.cat([score_classes, score_neg], dim=-1)
        return scores

    def get_accuracy(self, cls_score, labels):
        """Get custom accuracy w.r.t. cls_score and labels.

        Args:
            cls_score (torch.Tensor): The prediction with shape (N, C + 2).
            labels (torch.Tensor): The learning label of the prediction.

        Returns:
            Dict [str, torch.Tensor]: The accuracy for objectness and classes,
                 respectively.
        """
        pos_inds = labels < self.num_classes
        obj_labels = (labels == self.num_classes).long()
        cls_score_classes, cls_score_objectness = self._split_cls_score(
            cls_score)
        acc_objectness = accuracy(cls_score_objectness, obj_labels)
        acc_classes = accuracy(cls_score_classes[pos_inds], labels[pos_inds])
        acc = dict()
        acc['acc_objectness'] = acc_objectness
        acc['acc_classes'] = acc_classes
        return acc

    def forward(self,
                cls_score,
                labels,
                label_weights=None,
                avg_factor=None,
                reduction_override=None):
        """Forward function.

        Args:
            cls_score (torch.Tensor): The prediction with shape (N, C + 2).
            labels (torch.Tensor): The learning label of the prediction.
            label_weights (torch.Tensor, optional): Sample-wise loss weight.
            avg_factor (int, optional): Average factor that is used to average
                 the loss. Defaults to None.
            reduction (str, optional): The method used to reduce the loss.
                 Options are "none", "mean" and "sum".
        Returns:
            torch.Tensor | Dict [str, torch.Tensor]:
                 if return_dict == False: The calculated loss |
                 if return_dict == True: The dict of calculated losses
                 for objectness and classes, respectively.
        """
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        assert cls_score.size(-1) == self.num_classes + 2
        pos_inds = labels < self.num_classes
        # 0 for pos, 1 for neg
        obj_labels = (labels == self.num_classes).long()

        # accumulate the samples for each category
        unique_labels = labels.unique()
        for u_l in unique_labels:
            inds_ = labels == u_l.item()
            self.cum_samples[u_l] += inds_.sum()

        if label_weights is not None:
            label_weights = label_weights.float()
        else:
            label_weights = labels.new_ones(labels.size(), dtype=torch.float)

        cls_score_classes, cls_score_objectness = self._split_cls_score(
            cls_score)
        # calculate loss_cls_classes (only need pos samples)
        if pos_inds.sum() > 0:
            loss_cls_classes = self.loss_weight * self.cls_criterion(
                cls_score_classes[pos_inds], labels[pos_inds],
                label_weights[pos_inds], self.cum_samples[:self.num_classes],
                self.num_classes, self.p, self.q, self.eps, reduction,
                avg_factor)
        else:
            loss_cls_classes = cls_score_classes[pos_inds].sum()
        # calculate loss_cls_objectness
        loss_cls_objectness = self.loss_weight * cross_entropy(
            cls_score_objectness, obj_labels, label_weights, reduction,
            avg_factor)

        if self.return_dict:
            loss_cls = dict()
            loss_cls['loss_cls_objectness'] = loss_cls_objectness
            loss_cls['loss_cls_classes'] = loss_cls_classes
        else:
            loss_cls = loss_cls_classes + loss_cls_objectness
        return loss_cls



#  -*- coding: UTF-8 -*-
# @Time    : 18-3-26
# @File    : __init__.py
# @Author  : jian<jian@mltalker.com>
from .dataset import Dataset
from .cifar import Cifar10, Cifar100
from .pascal_voc import Pascal2007, Pascal2012
from .coco2017 import COCO2017
from .interhand26M import InterHand26M
from .lsp import LSP
from .visalso import VisalSO
from .imagenet import ImageNet
from .vggface import VGGFace
from .mnist import Mnist
from .flic import FLIC
from .lip import LIP
from .ade20k import ADE20K
from .lfw import LFW
from .cityscape import Cityscape
from .mpii import MPII

__all__ = [
  'Mnist', 'Cifar10', 'Cifar100', 'ImageNet', 'Pascal2007', 'Pascal2012', 'COCO2017', 'InterHand26M', 'LSP', 'VisalSO', 'VGGFace', 'FLIC', 'LIP', 'ADE20K', 'LFW', 'Cityscape', 'MPII', 'Dataset'
]



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch.nn.modules.module import Module
import math

def make_linear_layers(feat_dims, relu_final=True):
    layers = []
    for i in range(len(feat_dims)-1):
        layers.append(nn.Linear(feat_dims[i], feat_dims[i+1]))

        # Do not use ReLU for final estimation
        if i < len(feat_dims)-2 or (i == len(feat_dims)-2 and relu_final):
            layers.append(nn.ReLU(inplace=True))

    return nn.Sequential(*layers)

def make_conv_layers(feat_dims, kernel=3, stride=1, padding=1, bnrelu_final=True):
    layers = []
    for i in range(len(feat_dims)-1):
        layers.append(
            nn.Conv2d(
                in_channels=feat_dims[i],
                out_channels=feat_dims[i+1],
                kernel_size=kernel,
                stride=stride,
                padding=padding
                ))
        # Do not use BN and ReLU for final estimation
        if i < len(feat_dims)-2 or (i == len(feat_dims)-2 and bnrelu_final):
            layers.append(nn.BatchNorm2d(feat_dims[i+1]))
            layers.append(nn.ReLU(inplace=True))

    return nn.Sequential(*layers)

def make_deconv_layers(feat_dims, bnrelu_final=True):
    layers = []
    for i in range(len(feat_dims)-1):
        layers.append(
            nn.ConvTranspose2d(
                in_channels=feat_dims[i],
                out_channels=feat_dims[i+1],
                kernel_size=4,
                stride=2,
                padding=1,
                output_padding=0,
                bias=False))

        # Do not use BN and ReLU for final estimation
        if i < len(feat_dims)-2 or (i == len(feat_dims)-2 and bnrelu_final):
            layers.append(nn.BatchNorm2d(feat_dims[i+1]))
            layers.append(nn.ReLU(inplace=True))

    return nn.Sequential(*layers)

class Interpolate(nn.Module):
    def __init__(self, scale_factor, mode):
        super(Interpolate, self).__init__()
        self.interp = F.interpolate
        self.scale_factor = scale_factor
        self.mode = mode
        
    def forward(self, x):
        x = self.interp(x, scale_factor=self.scale_factor, mode=self.mode, align_corners=False)
        return x

def make_upsample_layers(feat_dims, bnrelu_final=True):
    layers = []
    for i in range(len(feat_dims)-1):
        layers.append(
            Interpolate(2, 'bilinear'))
        layers.append(
            nn.Conv2d(
                in_channels=feat_dims[i],
                out_channels=feat_dims[i+1],
                kernel_size=3,
                stride=1,
                padding=1
                ))

        # Do not use BN and ReLU for final estimation
        if i < len(feat_dims)-2 or (i == len(feat_dims)-2 and bnrelu_final):
            layers.append(nn.BatchNorm2d(feat_dims[i+1]))
            layers.append(nn.ReLU(inplace=True))

    return nn.Sequential(*layers)

class ResBlock(nn.Module):
    def __init__(self, in_feat, out_feat):
        super(ResBlock, self).__init__()
        self.in_feat = in_feat
        self.out_feat = out_feat

        self.conv = make_conv_layers([in_feat, out_feat, out_feat], bnrelu_final=False)
        self.bn = nn.BatchNorm2d(out_feat)
        if self.in_feat != self.out_feat:
            self.shortcut_conv = nn.Conv2d(in_feat,out_feat,kernel_size=1,stride=1,padding=0)
            self.shortcut_bn = nn.BatchNorm2d(out_feat)

    def forward(self, input):
        x = self.bn(self.conv(input))
        if self.in_feat != self.out_feat:
            x = F.relu(x + self.shortcut_bn(self.shortcut_conv(input)))
        else:
            x = F.relu(x + input)
        return x

def make_conv3d_layers(feat_dims, kernel=3, stride=1, padding=1, bnrelu_final=True):
    layers = []
    for i in range(len(feat_dims)-1):
        layers.append(
            nn.Conv3d(
                in_channels=feat_dims[i],
                out_channels=feat_dims[i+1],
                kernel_size=kernel,
                stride=stride,
                padding=padding
                ))
        # Do not use BN and ReLU for final estimation
        if i < len(feat_dims)-2 or (i == len(feat_dims)-2 and bnrelu_final):
            layers.append(nn.BatchNorm3d(feat_dims[i+1]))
            layers.append(nn.ReLU(inplace=True))

    return nn.Sequential(*layers)

def make_deconv3d_layers(feat_dims, bnrelu_final=True):
    layers = []
    for i in range(len(feat_dims)-1):
        layers.append(
            nn.ConvTranspose3d(
                in_channels=feat_dims[i],
                out_channels=feat_dims[i+1],
                kernel_size=4,
                stride=2,
                padding=1,
                output_padding=0,
                bias=False))

        # Do not use BN and ReLU for final estimation
        if i < len(feat_dims)-2 or (i == len(feat_dims)-2 and bnrelu_final):
            layers.append(nn.BatchNorm3d(feat_dims[i+1]))
            layers.append(nn.ReLU(inplace=True))

    return nn.Sequential(*layers)




# -*- coding: UTF-8 -*-
# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 14:44
# @File    : factory.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import os
import threading
from typing import Any, Dict, List, Tuple

from .execution.base_execution import *
from antgo.pipeline.hparam.hyperparameter import *
from antgo.pipeline.engine.operator_registry import *
from antgo.pipeline.engine.operator_loader import *


def op(operator_src: str,
       tag: str = 'main',
       arg: List[Any] = [],
       kwargs: Dict[str, Any] = {}):
  """
  Entry method which takes either operators tasks or paths to python files or class in notebook.
  An `Operator` object is created with the init args(kwargs).
  Args:
      operator_src (`str`):
          operators name or python file location or class in notebook.
      tag (`str`):
          Which tag to use for operators on hub, defaults to `main`.
  Returns
      (`typing.Any`)
          The `Operator` output.
  """
  if isinstance(operator_src, type):
    class_op = type('operators', (operator_src,), kwargs)
    return class_op.__new__(class_op, **kwargs)

  loader = OperatorLoader()
  return loader.load_operator(operator_src, arg, kwargs, tag)


class _OperatorLazyWrapper(  #
  BaseExecution):
  """
  operators wrapper for lazy initialization.
  """

  def __init__(self,
               real_name: str,
               index: Tuple[str],
               tag: str = 'main',
               arg: List[Any] = [],
               kws: Dict[str, Any] = {}) -> None:
    self._name = real_name.replace('.', '/').replace('_', '-')
    self._index = index
    self._tag = tag
    self._arg = arg
    self._kws = kws
    self._op = None
    self._lock = threading.Lock()
    self._op_config = self._kws.pop('op_config', None)
    # TODO: (How to apply such config)

  def __check_init__(self):
    with self._lock:
      if self._op is None:
        with param_scope(index=self._index):
          self._op = op(self._name,
                        self._tag,
                        arg=self._arg,
                        kwargs=self._kws)
          if hasattr(self._op, '__vcall__'):
            self.__has_vcall__ = True

  def get_op(self):
    self.__check_init__()
    return self._op

  @property
  def op_config(self):
    self.__check_init__()
    return self._op_config

  @property
  def function(self):
    return self._name

  @property
  def init_args(self):
    return self._kws

  @staticmethod
  def callback(real_name: str, index: Tuple[str], *arg, **kws):
    return _OperatorLazyWrapper(real_name, index, arg=arg, kws=kws)


@dynamic_dispatch
def ops(*arg, **kws):
    """
    Entry point for creating operators instances, for example:

    >>> op_instance = ops.my_namespace.my_repo_name(init_arg1=xxx, init_arg2=xxx)
    """

    # pylint: disable=protected-access
    with param_scope() as hp:
        real_name = hp._name
        index = hp._index
    return _OperatorLazyWrapper.callback(real_name, index, *arg, **kws)



# Copyright (c) OpenMMLab. All rights reserved.
from torch.autograd import Function
from torch.nn import functional as F


class SigmoidGeometricMean(Function):
    """Forward and backward function of geometric mean of two sigmoid
    functions.

    This implementation with analytical gradient function substitutes
    the autograd function of (x.sigmoid() * y.sigmoid()).sqrt(). The
    original implementation incurs none during gradient backprapagation
    if both x and y are very small values.
    """

    @staticmethod
    def forward(ctx, x, y):
        x_sigmoid = x.sigmoid()
        y_sigmoid = y.sigmoid()
        z = (x_sigmoid * y_sigmoid).sqrt()
        ctx.save_for_backward(x_sigmoid, y_sigmoid, z)
        return z

    @staticmethod
    def backward(ctx, grad_output):
        x_sigmoid, y_sigmoid, z = ctx.saved_tensors
        grad_x = grad_output * z * (1 - x_sigmoid) / 2
        grad_y = grad_output * z * (1 - y_sigmoid) / 2
        return grad_x, grad_y


sigmoid_geometric_mean = SigmoidGeometricMean.apply


def interpolate_as(source, target, mode='bilinear', align_corners=False):
    """Interpolate the `source` to the shape of the `target`.

    The `source` must be a Tensor, but the `target` can be a Tensor or a
    np.ndarray with the shape (..., target_h, target_w).

    Args:
        source (Tensor): A 3D/4D Tensor with the shape (N, H, W) or
            (N, C, H, W).
        target (Tensor | np.ndarray): The interpolation target with the shape
            (..., target_h, target_w).
        mode (str): Algorithm used for interpolation. The options are the
            same as those in F.interpolate(). Default: ``'bilinear'``.
        align_corners (bool): The same as the argument in F.interpolate().

    Returns:
        Tensor: The interpolated source Tensor.
    """
    assert len(target.shape) >= 2

    def _interpolate_as(source, target, mode='bilinear', align_corners=False):
        """Interpolate the `source` (4D) to the shape of the `target`."""
        target_h, target_w = target.shape[-2:]
        source_h, source_w = source.shape[-2:]
        if target_h != source_h or target_w != source_w:
            source = F.interpolate(
                source,
                size=(target_h, target_w),
                mode=mode,
                align_corners=align_corners)
        return source

    if len(source.shape) == 3:
        source = source[:, None, :, :]
        source = _interpolate_as(source, target, mode, align_corners)
        return source[:, 0, :, :]
    else:
        return _interpolate_as(source, target, mode, align_corners)



from antgo.dataflow.dataset.my_det_dataset import *
from antgo.dataflow.dataset.my_seg_dataset import *
from antgo.dataflow.imgaug.operators import *
from antgo.dataflow.imgaug.batch_operators import *
from antgo.dataflow.dataset.reader import *

# # 检测数据集
# dataset = CustomDetDataset('train', '/root/paddlejob/workspace/env_run/portrait/dataset/merge', {'anno_json': 'annot/valid.json'})

# # data_folder = '/Users/jian/Downloads/factory/dataset/MNIST'
# # dataset = Mnist('train', data_folder)
# reader = Reader(dataset, [
#     DecodeImage(),
#     RandomCrop(),
#     Resize((256,192)),
#     RandomDistort(),
#     Rotation(15),
#     RandomFlipImage(0.0),
#     NormalizeImage(is_channel_first=False,is_scale=False),
#     Permute(to_bgr=False, channel_first=True)
# ], [
#     Gt2TTFKPTTarget(num_classes=1, down_ratio=4),
#     PadBatch(pad_to_stride=16)
# ]
# , 4, True, True, inputs_def={
#     'fields': ['image']
# })

# for data in reader.iterator_value():
#   print(data)

# 分割数据集
dataset = CustomSegDataset('train', '/root/paddlejob/workspace/env_run/portrait/seg/humanseg_new', {'anno_file': 'train.txt'})
reader = Reader(dataset, [
    DecodeImage(),
    Rotation(80),
    Resize((256,192)),
    # RandomPaddingCrop(),
    Resize((128,128)),
    RandomFlipImage(),
    RandomDistort(),
    # NormalizeImage(is_channel_first=False,is_scale=False),
    Permute(to_bgr=False, channel_first=False)
], [
]
, 4, True, True, inputs_def={
    'fields': ['image','semantic']
}, use_process=True, worker_num=1)

for data in reader.iterator_value():
    image = data[0][0]
    mask = data[0][1]
    cv2.imwrite('./data.png', image)
    cv2.imwrite("./mask.png",(mask*255).astype(np.uint8))
    print('ss')



# Copyright (c) OpenMMLab. All rights reserved.
from .class_aware_sampler import ClassAwareSampler
from .distributed_sampler import DistributedSampler
from .group_sampler import DistributedGroupSampler, GroupSampler
from .infinite_sampler import InfiniteBatchSampler, InfiniteGroupBatchSampler
from .kv_sampler import KVSampler, DistributedKVSampler

__all__ = [
    'DistributedSampler', 'DistributedGroupSampler', 'GroupSampler',
    'InfiniteGroupBatchSampler', 'InfiniteBatchSampler', 'ClassAwareSampler',
    'KVSampler', 'DistributedKVSampler'
]



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 17:58
# @File    : nop.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function


from typing import Any, Dict, NamedTuple

from .base import PyOperator


class NOPOperator(PyOperator):
    """No-op operators. Input arguments are redefined as a `NamedTuple` and returned as
    outputs.
    """

    def __init__(self):
        #pylint: disable=useless-super-delegation
        super().__init__()

    def __call__(self, **args: Dict[str, Any]) -> NamedTuple:
        fields = [(name, type(val)) for name, val in args.items()]
        return NamedTuple('Outputs', fields)(**args)  # pylint: disable=not-callable



# -*- coding: UTF-8 -*-
# @Time    : 2022/5/2 19:41
# @File    : test_paddle_c.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import antgo.framework.paddle2torch.mapper as paddle
# import paddle
# import antgo.framework.paddle2torch.mapper.nn.functional as F

class LeNet(paddle.nn.Layer):
    def __init__(self):
        super(LeNet, self).__init__()
        self.conv1 = paddle.nn.Conv2D(in_channels=1, out_channels=6, kernel_size=5, stride=1, padding=2)
        self.max_pool1 = paddle.nn.MaxPool2D(kernel_size=2,  stride=2)
        self.conv2 = paddle.nn.Conv2D(in_channels=6, out_channels=16, kernel_size=5, stride=1)
        self.max_pool2 = paddle.nn.MaxPool2D(kernel_size=2, stride=2)
        self.linear1 = paddle.nn.Linear(in_features=16*5*5, out_features=120)
        self.linear2 = paddle.nn.Linear(in_features=120, out_features=84)
        self.linear3 = paddle.nn.Linear(in_features=84, out_features=10)

    def forward(self, x):
        x = self.conv1(x)
        x = paddle.nn.functional.relu(x)
        x = self.max_pool1(x)
        x = self.conv2(x)
        x = paddle.nn.functional.relu(x)
        x = self.max_pool2(x)
        x = paddle.flatten(x, start_axis=1,stop_axis=-1)
        x = self.linear1(x)
        x = paddle.nn.functional.relu(x)
        x = self.linear2(x)
        x = paddle.nn.functional.relu(x)
        x = self.linear3(x)
        return x

train_loader = paddle.io.DataLoader(train_dataset, batch_size=64, shuffle=True)
# 加载训练集 batch_size 设为 64
def train(model):
    model.train()
    epochs = 2
    optim = paddle.optimizer.Adam(learning_rate=0.001, parameters=model.parameters())
    # 用Adam作为优化函数
    for epoch in range(epochs):
        for batch_id, data in enumerate(train_loader()):
            x_data = data[0]
            y_data = data[1]
            predicts = model(x_data)
            loss = paddle.nn.functional.cross_entropy(predicts, y_data)
            # 计算损失
            acc = paddle.metric.accuracy(predicts, y_data)
            loss.backward()
            if batch_id % 300 == 0:
                print("epoch: {}, batch_id: {}, loss is: {}, acc is: {}".format(epoch, batch_id, loss.numpy(), acc.numpy()))
            optim.step()
            optim.clear_grad()

model = LeNet()
train(model)



"""Custom Operator."""
from .. import glue


def Custom(op_name):
    """Get custom operator whose name is `op_name`.

    Parameters
    ----------
    op_name: str
        The name of custom operator.

    Returns
    -------
    Custom Operator
    """
    assert op_name in glue.CUSTOM_OP_LIST, KeyError(
        'Operator {} not found'.format(op_name))
    return glue.CUSTOM_OP_LIST[op_name]


def CustomList():
    """Get the name list of custom operators
    Returns
    -------
    list of str
        The name list of custom operators
    """
    return glue.CUSTOM_OP_LIST.keys()



import collections.abc
import functools
import itertools
import subprocess
import warnings
from collections import abc
from importlib import import_module
from inspect import getfullargspec
from itertools import repeat
import glob
import os


# From PyTorch internals
def _ntuple(n):

    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return x
        return tuple(repeat(x, n))

    return parse


to_1tuple = _ntuple(1)
to_2tuple = _ntuple(2)
to_3tuple = _ntuple(3)
to_4tuple = _ntuple(4)
to_ntuple = _ntuple


def is_str(x):
    """Whether the input is an string instance.

    Note: This method is deprecated since python 2 is no longer supported.
    """
    return isinstance(x, str)


def import_modules_from_strings(imports, allow_failed_imports=False):
    """Import modules from the given list of strings.

    Args:
        imports (list | str | None): The given module names to be imported.
        allow_failed_imports (bool): If True, the failed imports will return
            None. Otherwise, an ImportError is raise. Default: False.

    Returns:
        list[module] | module | None: The imported modules.

    Examples:
        >>> osp, sys = import_modules_from_strings(
        ...     ['os.path', 'sys'])
        >>> import os.path as osp_
        >>> import sys as sys_
        >>> assert osp == osp_
        >>> assert sys == sys_
    """
    if not imports:
        return
    single_import = False
    if isinstance(imports, str):
        single_import = True
        imports = [imports]
    if not isinstance(imports, list):
        raise TypeError(
            f'custom_imports must be a list but got type {type(imports)}')
    imported = []
    for imp in imports:
        if not isinstance(imp, str):
            raise TypeError(
                f'{imp} is of type {type(imp)} and cannot be imported.')
        try:
            imported_tmp = import_module(imp)
        except ImportError:
            if allow_failed_imports:
                warnings.warn(f'{imp} failed to import and is ignored.',
                              UserWarning)
                imported_tmp = None
            else:
                raise ImportError
        imported.append(imported_tmp)
    if single_import:
        imported = imported[0]
    return imported


def iter_cast(inputs, dst_type, return_type=None):
    """Cast elements of an iterable object into some type.

    Args:
        inputs (Iterable): The input object.
        dst_type (type): Destination type.
        return_type (type, optional): If specified, the output object will be
            converted to this type, otherwise an iterator.

    Returns:
        iterator or specified type: The converted object.
    """
    if not isinstance(inputs, abc.Iterable):
        raise TypeError('inputs must be an iterable object')
    if not isinstance(dst_type, type):
        raise TypeError('"dst_type" must be a valid type')

    out_iterable = map(dst_type, inputs)

    if return_type is None:
        return out_iterable
    else:
        return return_type(out_iterable)


def list_cast(inputs, dst_type):
    """Cast elements of an iterable object into a list of some type.

    A partial method of :func:`iter_cast`.
    """
    return iter_cast(inputs, dst_type, return_type=list)


def tuple_cast(inputs, dst_type):
    """Cast elements of an iterable object into a tuple of some type.

    A partial method of :func:`iter_cast`.
    """
    return iter_cast(inputs, dst_type, return_type=tuple)


def is_seq_of(seq, expected_type, seq_type=None):
    """Check whether it is a sequence of some type.

    Args:
        seq (Sequence): The sequence to be checked.
        expected_type (type): Expected type of sequence items.
        seq_type (type, optional): Expected sequence type.

    Returns:
        bool: Whether the sequence is valid.
    """
    if seq_type is None:
        exp_seq_type = abc.Sequence
    else:
        assert isinstance(seq_type, type)
        exp_seq_type = seq_type
    if not isinstance(seq, exp_seq_type):
        return False
    for item in seq:
        if not isinstance(item, expected_type):
            return False
    return True


def is_list_of(seq, expected_type):
    """Check whether it is a list of some type.

    A partial method of :func:`is_seq_of`.
    """
    return is_seq_of(seq, expected_type, seq_type=list)


def is_tuple_of(seq, expected_type):
    """Check whether it is a tuple of some type.

    A partial method of :func:`is_seq_of`.
    """
    return is_seq_of(seq, expected_type, seq_type=tuple)


def slice_list(in_list, lens):
    """Slice a list into several sub lists by a list of given length.

    Args:
        in_list (list): The list to be sliced.
        lens(int or list): The expected length of each out list.

    Returns:
        list: A list of sliced list.
    """
    if isinstance(lens, int):
        assert len(in_list) % lens == 0
        lens = [lens] * int(len(in_list) / lens)
    if not isinstance(lens, list):
        raise TypeError('"indices" must be an integer or a list of integers')
    elif sum(lens) != len(in_list):
        raise ValueError('sum of lens and list length does not '
                         f'match: {sum(lens)} != {len(in_list)}')
    out_list = []
    idx = 0
    for i in range(len(lens)):
        out_list.append(in_list[idx:idx + lens[i]])
        idx += lens[i]
    return out_list


def concat_list(in_list):
    """Concatenate a list of list into a single list.

    Args:
        in_list (list): The list of list to be merged.

    Returns:
        list: The concatenated flat list.
    """
    return list(itertools.chain(*in_list))


def check_prerequisites(
        prerequisites,
        checker,
        msg_tmpl='Prerequisites "{}" are required in method "{}" but not '
        'found, please install them first.'):  # yapf: disable
    """A decorator factory to check if prerequisites are satisfied.

    Args:
        prerequisites (str of list[str]): Prerequisites to be checked.
        checker (callable): The checker method that returns True if a
            prerequisite is meet, False otherwise.
        msg_tmpl (str): The message template with two variables.

    Returns:
        decorator: A specific decorator.
    """

    def wrap(func):

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            requirements = [prerequisites] if isinstance(
                prerequisites, str) else prerequisites
            missing = []
            for item in requirements:
                if not checker(item):
                    missing.append(item)
            if missing:
                print(msg_tmpl.format(', '.join(missing), func.__name__))
                raise RuntimeError('Prerequisites not meet.')
            else:
                return func(*args, **kwargs)

        return wrapped_func

    return wrap


def _check_py_package(package):
    try:
        import_module(package)
    except ImportError:
        return False
    else:
        return True


def _check_executable(cmd):
    if subprocess.call(f'which {cmd}', shell=True) != 0:
        return False
    else:
        return True


def requires_package(prerequisites):
    """A decorator to check if some python packages are installed.

    Example:
        >>> @requires_package('numpy')
        >>> func(arg1, args):
        >>>     return numpy.zeros(1)
        array([0.])
        >>> @requires_package(['numpy', 'non_package'])
        >>> func(arg1, args):
        >>>     return numpy.zeros(1)
        ImportError
    """
    return check_prerequisites(prerequisites, checker=_check_py_package)


def requires_executable(prerequisites):
    """A decorator to check if some executable files are installed.

    Example:
        >>> @requires_executable('ffmpeg')
        >>> func(arg1, args):
        >>>     print(1)
        1
    """
    return check_prerequisites(prerequisites, checker=_check_executable)


def deprecated_api_warning(name_dict, cls_name=None):
    """A decorator to check if some arguments are deprecate and try to replace
    deprecate src_arg_name to dst_arg_name.

    Args:
        name_dict(dict):
            key (str): Deprecate argument names.
            val (str): Expected argument names.

    Returns:
        func: New function.
    """

    def api_warning_wrapper(old_func):

        @functools.wraps(old_func)
        def new_func(*args, **kwargs):
            # get the arg spec of the decorated method
            args_info = getfullargspec(old_func)
            # get name of the function
            func_name = old_func.__name__
            if cls_name is not None:
                func_name = f'{cls_name}.{func_name}'
            if args:
                arg_names = args_info.args[:len(args)]
                for src_arg_name, dst_arg_name in name_dict.items():
                    if src_arg_name in arg_names:
                        warnings.warn(
                            f'"{src_arg_name}" is deprecated in '
                            f'`{func_name}`, please use "{dst_arg_name}" '
                            'instead', DeprecationWarning)
                        arg_names[arg_names.index(src_arg_name)] = dst_arg_name
            if kwargs:
                for src_arg_name, dst_arg_name in name_dict.items():
                    if src_arg_name in kwargs:

                        assert dst_arg_name not in kwargs, (
                            f'The expected behavior is to replace '
                            f'the deprecated key `{src_arg_name}` to '
                            f'new key `{dst_arg_name}`, but got them '
                            f'in the arguments at the same time, which '
                            f'is confusing. `{src_arg_name} will be '
                            f'deprecated in the future, please '
                            f'use `{dst_arg_name}` instead.')

                        warnings.warn(
                            f'"{src_arg_name}" is deprecated in '
                            f'`{func_name}`, please use "{dst_arg_name}" '
                            'instead', DeprecationWarning)
                        kwargs[dst_arg_name] = kwargs.pop(src_arg_name)

            # apply converted arguments to the decorated method
            output = old_func(*args, **kwargs)
            return output

        return new_func

    return api_warning_wrapper


def is_method_overridden(method, base_class, derived_class):
    """Check if a method of base class is overridden in derived class.

    Args:
        method (str): the method name to check.
        base_class (type): the class of the base class.
        derived_class (type | Any): the class or instance of the derived class.
    """
    assert isinstance(base_class, type), \
        "base_class doesn't accept instance, Please pass class instead."

    if not isinstance(derived_class, type):
        derived_class = derived_class.__class__

    base_method = getattr(base_class, method)
    derived_method = getattr(derived_class, method)
    return derived_method != base_method


def has_method(obj: object, method: str) -> bool:
    """Check whether the object has a method.

    Args:
        method (str): The method name to check.
        obj (object): The object to check.

    Returns:
        bool: True if the object has the method else False.
    """
    return hasattr(obj, method) and callable(getattr(obj, method))


def find_latest_checkpoint(path, suffix='pth'):
    """Find the latest checkpoint from the working directory.

    Args:
        path(str): The path to find checkpoints.
        suffix(str): File extension.
            Defaults to pth.

    Returns:
        latest_path(str | None): File path of the latest checkpoint.
    References:
        .. [1] https://github.com/microsoft/SoftTeacher
                  /blob/main/ssod/utils/patch.py
    """
    if not osp.exists(path):
        warnings.warn('The path of checkpoints does not exist.')
        return None
    if osp.exists(osp.join(path, f'latest.{suffix}')):
        return osp.join(path, f'latest.{suffix}')

    checkpoints = glob.glob(osp.join(path, f'*.{suffix}'))
    if len(checkpoints) == 0:
        warnings.warn('There are no checkpoints in the path.')
        return None
    latest = -1
    latest_path = None
    for checkpoint in checkpoints:
        count = int(osp.basename(checkpoint).split('_')[-1].split('.')[0])
        if count > latest:
            latest = count
            latest_path = checkpoint
    return latest_path


def update_data_root(cfg, logger=None):
    """Update data root according to env MMDET_DATASETS.

    If set env ANTGO_DATASETS, update cfg.data_root according to
    ANTGO_DATASETS. Otherwise, using cfg.data_root as default.

    Args:
        cfg (antgo.framework.helper.utils.Config): The model config need to modify
        logger (logging.Logger | str | None): the way to print msg
    """
    # assert isinstance(cfg, antgo.framework.helper.utils.Config), \
    #     f'cfg got wrong type: {type(cfg)}, antgo.framework.helper.utils.Config'

    if 'ANTGO_DATASETS' in os.environ:
        dst_root = os.environ['ANTGO_DATASETS']
        print_log(f'MMDET_DATASETS has been set to be {dst_root}.'
                  f'Using {dst_root} as data root.')
    else:
        return

    # assert isinstance(cfg,antgo.framework.helper.utils.Config), \
    #     f'cfg got wrong type: {type(cfg)}, expected antgo.framework.helper.utils.Config'

    def update(cfg, src_str, dst_str):
        for k, v in cfg.items():
            # if isinstance(v, antgo.framework.helper.utils.ConfigDict):
            #     update(cfg[k], src_str, dst_str)
            if isinstance(v, str) and src_str in v:
                cfg[k] = v.replace(src_str, dst_str)
            else:
                 update(cfg[k], src_str, dst_str)

    update(cfg.data, cfg.data_root, dst_root)
    cfg.data_root = dst_root



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/13 21:05
# @File    : util.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.pipeline.models.utils.common import *


class Detect(nn.Module):
  stride = None  # strides computed during build
  def __init__(self, nc=80, anchors=(), ch=()):  # detection layer
    super(Detect, self).__init__()
    self.nc = nc  # number of classes
    self.no = nc + 5  # number of outputs per anchor
    self.nl = len(anchors)  # number of detection layers
    self.na = len(anchors[0]) // 2  # number of anchors
    self.grid = [torch.zeros(1)] * self.nl  # init grid
    a = torch.tensor(anchors).float().view(self.nl, -1, 2)
    self.register_buffer('anchors', a)  # shape(nl,na,2)
    self.register_buffer('anchor_grid', a.clone().view(self.nl, 1, -1, 1, 1, 2))  # shape(nl,1,na,1,1,2)
    self.m = nn.ModuleList(nn.Conv2d(x, self.no * self.na, 1) for x in ch)  # output conv

  def forward(self, x):
    # x = x.copy()  # for profiling
    z = []  # inference output
    for i in range(self.nl):
      x[i] = self.m[i](x[i])  # conv
      bs, _, ny, nx = x[i].shape  # x(bs,255,20,20) to x(bs,3,20,20,85)
      x[i] = x[i].view(bs, self.na, self.no, ny, nx).permute(0, 1, 3, 4, 2).contiguous()

      if self.grid[i].shape[2:4] != x[i].shape[2:4]:
        self.grid[i] = self._make_grid(nx, ny).to(x[i].device)
      y = x[i].sigmoid()

      y[..., 0:2] = (y[..., 0:2] * 2. - 0.5 + self.grid[i]) * self.stride[i]  # xy
      y[..., 2:4] = (y[..., 2:4] * 2) ** 2 * self.anchor_grid[i]  # wh
      z.append(y.view(bs, -1, self.no))

    out = (torch.cat(z, 1), x)
    return out

  @staticmethod
  def _make_grid(nx=20, ny=20):
    yv, xv = torch.meshgrid([torch.arange(ny), torch.arange(nx)])
    return torch.stack((xv, yv), 2).view((1, 1, ny, nx, 2)).float()

  def convert(self, z):
    z = torch.cat(z, 1)
    box = z[:, :, :4]
    conf = z[:, :, 4:5]
    score = z[:, :, 5:]
    score *= conf
    convert_matrix = torch.tensor([[1, 0, 1, 0], [0, 1, 0, 1], [-0.5, 0, 0.5, 0], [0, -0.5, 0, 0.5]],
                                  dtype=torch.float32,
                                  device=z.device)
    box @= convert_matrix
    return (box, score)


def yolo_parse_model(d, ch):  # model_dict, input_channels(3)
  # logger.info('\n%3s%18s%3s%10s  %-40s%-30s' % ('', 'from', 'n', 'params', 'module', 'arguments'))
  anchors, nc, gd, gw = d['anchors'], d['nc'], d['depth_multiple'], d['width_multiple']
  na = (len(anchors[0]) // 2) if isinstance(anchors, list) else anchors  # number of anchors
  no = na * (nc + 5)  # number of outputs = anchors * (classes + 5)

  layers, save, c2 = [], [], ch[-1]  # layers, savelist, ch out
  for i, (f, n, m, args) in enumerate(d['backbone'] + d['head']):  # from, number, module, args
    m = eval(m) if isinstance(m, str) else m  # eval strings
    for j, a in enumerate(args):
      try:
        args[j] = eval(a) if isinstance(a, str) else a  # eval strings
      except:
        pass

    n = max(round(n * gd), 1) if n > 1 else n  # depth gain
    if m in [nn.Conv2d, Conv, RobustConv, RobustConv2, DWConv, GhostConv, RepConv, RepConv_OREPA, DownC,
             SPP, SPPF, SPPCSPC, GhostSPPCSPC, MixConv2d, Focus, Stem, GhostStem, CrossConv,
             Bottleneck, BottleneckCSPA, BottleneckCSPB, BottleneckCSPC,
             RepBottleneck, RepBottleneckCSPA, RepBottleneckCSPB, RepBottleneckCSPC,
             Res, ResCSPA, ResCSPB, ResCSPC,
             RepRes, RepResCSPA, RepResCSPB, RepResCSPC,
             ResX, ResXCSPA, ResXCSPB, ResXCSPC,
             RepResX, RepResXCSPA, RepResXCSPB, RepResXCSPC,
             Ghost, GhostCSPA, GhostCSPB, GhostCSPC,
             SwinTransformerBlock, STCSPA, STCSPB, STCSPC,
             SwinTransformer2Block, ST2CSPA, ST2CSPB, ST2CSPC]:
      c1, c2 = ch[f], args[0]
      if c2 != no:  # if not output
        c2 = make_divisible(c2 * gw, 8)

      args = [c1, c2, *args[1:]]
      if m in [DownC, SPPCSPC, GhostSPPCSPC,
               BottleneckCSPA, BottleneckCSPB, BottleneckCSPC,
               RepBottleneckCSPA, RepBottleneckCSPB, RepBottleneckCSPC,
               ResCSPA, ResCSPB, ResCSPC,
               RepResCSPA, RepResCSPB, RepResCSPC,
               ResXCSPA, ResXCSPB, ResXCSPC,
               RepResXCSPA, RepResXCSPB, RepResXCSPC,
               GhostCSPA, GhostCSPB, GhostCSPC,
               STCSPA, STCSPB, STCSPC,
               ST2CSPA, ST2CSPB, ST2CSPC]:
        args.insert(2, n)  # number of repeats
        n = 1
    elif m is nn.BatchNorm2d:
      args = [ch[f]]
    elif m is Concat:
      c2 = sum([ch[x] for x in f])
    elif m is Chuncat:
      c2 = sum([ch[x] for x in f])
    elif m is Shortcut:
      c2 = ch[f[0]]
    elif m is Foldcut:
      c2 = ch[f] // 2
    elif m in [Detect]:
      args.append([ch[x] for x in f])
      if isinstance(args[1], int):  # number of anchors
        args[1] = [list(range(args[1] * 2))] * len(f)
    elif m is ReOrg:
      c2 = ch[f] * 4
    elif m is Contract:
      c2 = ch[f] * args[0] ** 2
    elif m is Expand:
      c2 = ch[f] // args[0] ** 2
    else:
      c2 = ch[f]

    m_ = nn.Sequential(*[m(*args) for _ in range(n)]) if n > 1 else m(*args)  # module
    t = str(m)[8:-2].replace('__main__.', '')  # module type
    np = sum([x.numel() for x in m_.parameters()])  # number params
    m_.i, m_.f, m_.type, m_.np = i, f, t, np  # attach index, 'from' index, type, number params
    # logger.info('%3s%18s%3s%10.0f  %-40s%-30s' % (i, f, n, np, t, args))  # print
    save.extend(x % i for x in ([f] if isinstance(f, int) else f) if x != -1)  # append to savelist
    layers.append(m_)
    if i == 0:
      ch = []
    ch.append(c2)
  return nn.Sequential(*layers), sorted(save)




# Copyright (c) OpenMMLab. All rights reserved.

# This script consists of several convert functions which
# can modify the weights of model in original repo to be
# pre-trained weights.

from collections import OrderedDict

import torch


def pvt_convert(ckpt):
    new_ckpt = OrderedDict()
    # Process the concat between q linear weights and kv linear weights
    use_abs_pos_embed = False
    use_conv_ffn = False
    for k in ckpt.keys():
        if k.startswith('pos_embed'):
            use_abs_pos_embed = True
        if k.find('dwconv') >= 0:
            use_conv_ffn = True
    for k, v in ckpt.items():
        if k.startswith('head'):
            continue
        if k.startswith('norm.'):
            continue
        if k.startswith('cls_token'):
            continue
        if k.startswith('pos_embed'):
            stage_i = int(k.replace('pos_embed', ''))
            new_k = k.replace(f'pos_embed{stage_i}',
                              f'layers.{stage_i - 1}.1.0.pos_embed')
            if stage_i == 4 and v.size(1) == 50:  # 1 (cls token) + 7 * 7
                new_v = v[:, 1:, :]  # remove cls token
            else:
                new_v = v
        elif k.startswith('patch_embed'):
            stage_i = int(k.split('.')[0].replace('patch_embed', ''))
            new_k = k.replace(f'patch_embed{stage_i}',
                              f'layers.{stage_i - 1}.0')
            new_v = v
            if 'proj.' in new_k:
                new_k = new_k.replace('proj.', 'projection.')
        elif k.startswith('block'):
            stage_i = int(k.split('.')[0].replace('block', ''))
            layer_i = int(k.split('.')[1])
            new_layer_i = layer_i + use_abs_pos_embed
            new_k = k.replace(f'block{stage_i}.{layer_i}',
                              f'layers.{stage_i - 1}.1.{new_layer_i}')
            new_v = v
            if 'attn.q.' in new_k:
                sub_item_k = k.replace('q.', 'kv.')
                new_k = new_k.replace('q.', 'attn.in_proj_')
                new_v = torch.cat([v, ckpt[sub_item_k]], dim=0)
            elif 'attn.kv.' in new_k:
                continue
            elif 'attn.proj.' in new_k:
                new_k = new_k.replace('proj.', 'attn.out_proj.')
            elif 'attn.sr.' in new_k:
                new_k = new_k.replace('sr.', 'sr.')
            elif 'mlp.' in new_k:
                string = f'{new_k}-'
                new_k = new_k.replace('mlp.', 'ffn.layers.')
                if 'fc1.weight' in new_k or 'fc2.weight' in new_k:
                    new_v = v.reshape((*v.shape, 1, 1))
                new_k = new_k.replace('fc1.', '0.')
                new_k = new_k.replace('dwconv.dwconv.', '1.')
                if use_conv_ffn:
                    new_k = new_k.replace('fc2.', '4.')
                else:
                    new_k = new_k.replace('fc2.', '3.')
                string += f'{new_k} {v.shape}-{new_v.shape}'
        elif k.startswith('norm'):
            stage_i = int(k[4])
            new_k = k.replace(f'norm{stage_i}', f'layers.{stage_i - 1}.2')
            new_v = v
        else:
            new_k = k
            new_v = v
        new_ckpt[new_k] = new_v

    return new_ckpt


def swin_converter(ckpt):

    new_ckpt = OrderedDict()

    def correct_unfold_reduction_order(x):
        out_channel, in_channel = x.shape
        x = x.reshape(out_channel, 4, in_channel // 4)
        x = x[:, [0, 2, 1, 3], :].transpose(1,
                                            2).reshape(out_channel, in_channel)
        return x

    def correct_unfold_norm_order(x):
        in_channel = x.shape[0]
        x = x.reshape(4, in_channel // 4)
        x = x[[0, 2, 1, 3], :].transpose(0, 1).reshape(in_channel)
        return x

    for k, v in ckpt.items():
        if k.startswith('head'):
            continue
        elif k.startswith('layers'):
            new_v = v
            if 'attn.' in k:
                new_k = k.replace('attn.', 'attn.w_msa.')
            elif 'mlp.' in k:
                if 'mlp.fc1.' in k:
                    new_k = k.replace('mlp.fc1.', 'ffn.layers.0.0.')
                elif 'mlp.fc2.' in k:
                    new_k = k.replace('mlp.fc2.', 'ffn.layers.1.')
                else:
                    new_k = k.replace('mlp.', 'ffn.')
            elif 'downsample' in k:
                new_k = k
                if 'reduction.' in k:
                    new_v = correct_unfold_reduction_order(v)
                elif 'norm.' in k:
                    new_v = correct_unfold_norm_order(v)
            else:
                new_k = k
            new_k = new_k.replace('layers', 'stages', 1)
        elif k.startswith('patch_embed'):
            new_v = v
            if 'proj' in k:
                new_k = k.replace('proj', 'projection')
            else:
                new_k = k
        else:
            new_v = v
            new_k = k

        new_ckpt['backbone.' + new_k] = new_v

    return new_ckpt



from pydoc import classname
import torch
import torch.nn.functional as F
from antgo.framework.helper.models.builder import MODELS, build_model
from antgo.framework.helper.multi_stream_module import MultiSteamModule
from antgo.framework.helper.models.ema_module import *
from antgo.framework.helper.models.builder import HEADS, build_loss
from antgo.framework.helper.models.detectors.utils.gaussian_target import gaussian2D
import cv2
import numpy as np


@MODELS.register_module()
class DETMPL(MultiSteamModule):
    def __init__(self, model: dict, ema=0.0, train_cfg=None, test_cfg=None) -> None:
        if test_cfg is None:
            test_cfg = dict()
        test_cfg.update(
            {
                'inference_on': 'student'
            }
        )
        super(DETMPL, self).__init__(
            dict(
                teacher=build_model(model['teacher']),
                student=build_model(model['student']),
            ),
            train_cfg=train_cfg if train_cfg is not None else dict(),
            test_cfg=test_cfg if test_cfg is not None else dict(),
        )

        self.avg_student = None
        if ema > 0:
            self.avg_student = ModelEMA(self.student, ema)
            self.inference_on = self.avg_student

        self.temperature = train_cfg.get('temperature', 1.0)
        self.lambda_u = train_cfg.get('lambda_u', 1.0)
        self.uda_steps = train_cfg.get('uda_steps', 1)
        self.grad_clip = train_cfg.get('grad_clip', 1e9)
        self.ema = train_cfg.get('ema', 0)
        self.threshold = train_cfg.get('threshold', 0.95)
        self.label_batch_size = train_cfg.get('label_batch_size', 5)
        self.unlabel_batch_size = train_cfg.get('unlabel_batch_size', 3)    # 无标签数据量 在一个batch里
        
        self.use_sigmoid = train_cfg.get('use_sigmoid', True)               # 是否将挑选出来的featuremap 使用sigmoid 
        self.positive_threshold = train_cfg.get('positive_threshold', 0.8)
        self.negative_threshold = train_cfg.get('negative_threshold', 0.1)
        loss_ch = train_cfg.get('loss_ch', dict(type='GaussianFocalLoss', loss_weight=1.0))
        self.loss_center_heatmap = build_loss(loss_ch)
        
        # 生成gaussian 伪标签的高斯核
        gaussian_kernel = gaussian2D(2, sigma=1)
        gaussian_kernel_h, gaussian_kernel_w = gaussian_kernel.shape[:2]
        self.gaussian_kernel = torch.reshape(gaussian_kernel, (1,1,gaussian_kernel_h, gaussian_kernel_w))
        
    def uda_loss(self, logits_uw, logits_us):
        # logits_uw: 弱增强 结果
        # logits_us: 强增强 结果
        mask, soft_pseudo_label, hard_pseudo_label = self.make_pseudo_label(logits_uw)

        # 损失
        logits_us = logits_us.sigmoid()
        loss_u =  F.binary_cross_entropy(logits_us, soft_pseudo_label, reduction="none") * mask
        loss_u = logits_us.sum()/(mask.sum()+1e-4)
        
        return mask, hard_pseudo_label, loss_u

    def train_step(self, data, optimizer, **kwargs):
        image = data.pop('image')
        other = data
        label_batch_size = self.label_batch_size
        batch_size = image.shape[0]
        unlabel_batch_size = (batch_size - label_batch_size)//2

        label_images, unlabel_ws_images = \
            torch.split(image, [label_batch_size, batch_size-label_batch_size],dim=0) 

        unlabel_w_images=torch.index_select(unlabel_ws_images,dim=0,index=torch.tensor([i for i in range(0,2*unlabel_batch_size,2)]))
        unlabel_s_images=torch.index_select(unlabel_ws_images,dim=0,index=torch.tensor([i for i in range(1,2*unlabel_batch_size,2)]))

        label_kwargs = {}
        unlabel_w_kwargs = {}
        unlabel_s_kwargs = {}
        for k, v in other.items():
            if isinstance(v, list):
                v_label_kwargs = v[:label_batch_size]
                v_unlabel_ws_kwargs = v[label_batch_size:]
                v_unlabel_w_kwargs = [v_unlabel_ws_kwargs[i] for i in range(0, 2*unlabel_batch_size, 2)]
                v_unlabel_s_kwargs = [v_unlabel_ws_kwargs[i] for i in range(1, 2*unlabel_batch_size, 2)]
                
                label_kwargs[k] = v_label_kwargs
                unlabel_w_kwargs[k] = v_unlabel_w_kwargs
                unlabel_s_kwargs[k] = v_unlabel_s_kwargs
                continue

            v_label_kwargs, v_unlabel_ws_kwargs = \
                torch.split(v, [label_batch_size, unlabel_batch_size * 2],dim=0) 
            label_kwargs[k] = v_label_kwargs

            v_unlabel_w_kwargs=torch.index_select(v_unlabel_ws_kwargs,dim=0,index=torch.tensor([i for i in range(0,2*unlabel_batch_size,2)]))
            v_unlabel_s_kwargs=torch.index_select(v_unlabel_ws_kwargs,dim=0,index=torch.tensor([i for i in range(1,2*unlabel_batch_size,2)]))

            unlabel_w_kwargs[k] = v_unlabel_w_kwargs
            unlabel_s_kwargs[k] = v_unlabel_s_kwargs

        t_images = torch.cat((label_images, unlabel_w_images, unlabel_s_images))
        t_kwargs = {k: None for k in label_kwargs.keys()}
        # image_meta 标配信息
        t_kwargs['image_meta'] = []
        t_kwargs['image_meta'].extend(label_kwargs['image_meta'])
        t_kwargs['image_meta'].extend(unlabel_w_kwargs['image_meta'])
        t_kwargs['image_meta'].extend(unlabel_s_kwargs['image_meta'])        

        # 每个模型一般由两个模块组成 forward_train -> loss
        t_logits, t_reg = self.teacher.forward_train(t_images, **t_kwargs) # tuple
        
        t_logits_l = t_logits[:label_batch_size]
        t_logits_uw, t_logits_us = t_logits[label_batch_size:].chunk(2)
        
        t_reg_l = t_reg[:label_batch_size]
        t_reg_uw, t_reg_us = t_reg[label_batch_size:].chunk(2)
        del t_logits, t_reg

        # teacher loss in label data
        t_loss_dict = self.teacher.loss(t_logits_l, t_reg_l, **label_kwargs)
        t_loss_l = sum(_value for _, _value in t_loss_dict.items())

        # teacher loss in unlabel data
        mask, hard_pseudo_label, t_loss_u = self.uda_loss(t_logits_uw.detach() / self.temperature, t_logits_us)

        weight_u = self.lambda_u * min(1., (kwargs['iter'] + 1) / self.uda_steps)
        t_loss_uda = t_loss_l + weight_u * t_loss_u

        s_images = torch.cat((label_images, unlabel_s_images))
        s_kwargs = {k: None for k in label_kwargs.keys()}
        s_kwargs['image_meta'] = []
        s_kwargs['image_meta'].extend(label_kwargs['image_meta'])
        s_kwargs['image_meta'].extend(unlabel_s_kwargs['image_meta'])
        s_logits, s_reg = self.student.forward_train(s_images, **s_kwargs)
        s_logits_l = s_logits[:label_batch_size]
        s_logits_us = s_logits[label_batch_size:]
        
        s_reg_l = s_reg[:label_batch_size]
        s_reg_us = s_reg[label_batch_size:]
        del s_logits, s_reg

        # 查看student未使用伪标签更新前，在有标签数据上的损失
        s_loss_dict_old = self.student.loss(s_logits_l.detach(), s_reg_l.detach(), **label_kwargs)
        s_loss_l_old = sum(_value for _, _value in s_loss_dict_old.items())
        
        # student使用伪标签计算损失并更新 
        # TODO，加入mask，消除不确定区域影响
        s_loss = self.loss_center_heatmap(s_logits_us.sigmoid(), hard_pseudo_label)        
        s_loss.backward()
        if self.grad_clip > 0:
            nn.utils.clip_grad_norm_(self.student.parameters(), self.grad_clip)
        
        optimizer['student'].step()
        if self.ema > 0:
            self.avg_student.update_parameters(self.student)     

        ####
        # 查看student在使用伪标签更新后，在有标签数据上的损失
        with torch.no_grad():
            s_loss_dict_new = self.student.forward_train(label_images, **label_kwargs)
            s_loss_l_new = sum(_value for _, _value in s_loss_dict_new.items())
 
        # theoretically correct formula (https://github.com/kekmodel/MPL-pytorch/issues/6)
        # dot_product = s_loss_l_old - s_loss_l_new

        # author's code formula
        dot_product = s_loss_l_new - s_loss_l_old
        # moving_dot_product = moving_dot_product * 0.99 + dot_product * 0.01
        # dot_product = dot_product - moving_dot_product

        # TODO，这里感觉应该改成使用弱增强来获得伪标签
        # _, _, hard_pseudo_label = self.make_pseudo_label(t_logits_us.detach())
        # 使用来自于student的反馈信号+teacher的硬标签损失
        t_loss_mpl = dot_product * self.loss_center_heatmap(t_logits_us.sigmoid(), hard_pseudo_label)
        # test
        # t_loss_mpl = torch.tensor(0.).to(args.device)
        t_loss = t_loss_uda + t_loss_mpl
        t_loss.backward()
        if self.grad_clip > 0:
            nn.utils.clip_grad_norm_(self.teacher.parameters(), self.grad_clip)
        
        optimizer['teacher'].step()
        self.teacher.zero_grad()
        self.student.zero_grad()

        losses = {
            's_loss': s_loss,
            't_loss': t_loss,
            't_loss_l': t_loss_l,
            't_loss_u': t_loss_u,
            't_loss_mpl': t_loss_mpl,            
        }

        losses, log_vars = self._parse_losses(losses)
        outputs = dict(
            loss=losses, log_vars=log_vars, num_samples=batch_size)

        outputs['log_vars'].update(
            {'mask': mask.mean().cpu().item()}
        )
        return outputs

    def make_pseudo_label(self, logits_u):
        # 返回 mask, soft_pseudo_label, hard_pseudo_label
        # NxCxHxW
        # logits_u (已经是概率 语义)
        # 计算软标签
        soft_pseudo_label = logits_u.sigmoid()
            
        # 计算有效区域
        # 小于低阈值区域为绝对背景
        # 高于高阈值区域为绝对前景
        positive_mask = soft_pseudo_label.ge(self.positive_threshold).float()
        negative_mask = soft_pseudo_label.le(self.negative_threshold).float()
        mask = ((positive_mask + negative_mask) > 0.5).float()
        
        # 计算硬标签
        hard_pseudo_label = torch.max_pool2d(soft_pseudo_label, 3, 1, padding=1)
        hard_pseudo_label = (hard_pseudo_label == soft_pseudo_label).float()        # 最大值位置为1，其余位置为0

        batch_size, class_num = logits_u.shape[:2]
        with torch.no_grad():
            weight = self.gaussian_kernel.repeat(class_num, 1,1,1)
            hard_pseudo_label = F.conv2d(hard_pseudo_label, weight, bias=None, stride=1, padding=5//2, groups=class_num)

            # 调整最大值为1
            max_val= torch.max(hard_pseudo_label.view(batch_size, class_num, -1), dim=-1, keepdim=True)[0]
            hard_pseudo_label = hard_pseudo_label / (max_val.view(batch_size, class_num, 1, 1)+1e-10)

        return mask, soft_pseudo_label, hard_pseudo_label

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs):
        # 同时兼容从base模型和dense模型加载
        if not any(["student" in key or "teacher" in key for key in state_dict.keys()]):
            keys = list(state_dict.keys())
            state_dict.update({"teacher." + k: state_dict[k] for k in keys})
            state_dict.update({"student." + k: state_dict[k] for k in keys})
            for k in keys:
                state_dict.pop(k)

        return super()._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )


# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import paddle


def device_count():
    gpu_useful = paddle.get_device().startswith("gpu")
    if gpu_useful:
        device_str = os.environ["CUDA_VISIBLE_DEVICES"]
        seg = device_str.split(",")
        return len(seg)
    else:
        return 0



# encoding=utf-8
# @Time    : 23-03-22
# @File    : client.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from socket import SOCK_STREAM
import sys
import os
from unicodedata import name
from antvis.client.httprpc import *
from antgo.framework.helper.fileio import *
from antgo import config
from antgo import script
from antgo.utils.args import *
from antgo.framework.helper.tools.util import *
import time
from queue import PriorityQueue
from socketserver import StreamRequestHandler as Tcp
import socketserver
from socket import *
import threading
import shutil
import copy


'''
后台常住,需要定时获取项目最新进展 

'''
server_manager = None
file_lock = threading.Lock()

class BackgroundTCP(Tcp):
    def handle(self):
        while True:
            data = self.request.recv(1024)
            if not data:
                break

            data = data.decode('utf-8')
            if data == 'ping':
                content = {'status': True, 'message': 'ping recv'}
                self.request.sendall(json.dumps(content).encode('utf-8'))
                break

            try:
                global server_manager
                status = server_manager.trigger(data)
                content = {'status': status, 'message': ''}
                self.request.sendall(json.dumps(content).encode('utf-8'))
            except:
                logging.error(f'background server recv abnormal data {data}')

            break

def exp_basic_info():
    return {
        "exp": "", 
        "id": "",
        "branch": "", 
        "commit": "", 
        "metric": {}, 
        "dataset": {"test": "", "train": []},
        "checkpoint": "", 
        "create_time": time.strftime('%Y-%m-%dx%H-%M-%S',time.localtime(time.time())), 
        "finish_time": "",
        "state": "",    # running, finish, stop, default
        "stage": "",    # which stage of progress
    }


def dataset_basic_info():
    return {
        "tag": "",
        "num": 0,
        "status": True,
        'address': ""
    }

class ServerBase(object):
    def __init__(self, *args, **kwargs) -> None:
        super(ServerBase, self).__init__()
        self.root = kwargs.get('root', './')
        self.ext_module = kwargs.get('ext_module', None)
        self.timer = 10         # 10分钟一次监控, debug 10*60
        self.task_queue = PriorityQueue()
        self.task_set = set()       # 用于校验是否存在同类型任务

        # supervised            优先级2
        # semi-supervised       优先级3
        # distillation          优先级3
        # activelearning        优先级1
        # self.task_order = [('activelearning', 'label'),('supervised', 'activelearning'),('label', 'supervised'),('supervised','semi-supervised'),('supervised','distillation')]
        self.task_order = [('activelearning', 'label'),('supervised', 'activelearning'), ('label', 'supervised')]

        # self.task_order = [('activelearning', 'label'),('supervised', 'activelearning'),('label', 'supervised')]
        # self.task_order = [('supervised','semi-supervised')]
        self.task_priority = {'supervised':1, 'activelearning': 2, 'semi-supervised':3, 'distillation':3, 'label': 1}
        self.task_cmd = {
            "activelearning": "antgo activelearning",
            "label": "antgo tool label/start",
            "supervised": "antgo train",
            "semi-supervised": "antgo train",
            "distillation": "antgo train"
        }

        # 启动定时任务
        periodical_func = threading.Timer(self.timer, self.watch)
        periodical_func.start()

    def submit(self):
        raise NotImplementedError

    def watch(self):
        raise NotImplementedError

    def clear(self):
        self.task_queue.clear()

    def trigger(self, event):
        # 当开发者主动更新 1. 标签数据; 2. 无标签数据; 3. 更换product模型; 触发
        file_lock.acquire()
        project_name, project_event = event.split('.')
        project_info = {}
        if os.path.exists(os.path.join(config.AntConfig.task_factory, f'{project_name}.json')):
            with open(os.path.join(config.AntConfig.task_factory, f'{project_name}.json'), 'r') as fp:
                project_info = json.load(fp)        
    
        if len(project_info) == 0:
            logging.warn(f'Missing {project_name} project info.')
            file_lock.release()            
            return False

        if project_info['product'] == '' or project_info['product'] not in project_info['exp']:
            logging.warn(f'Missing project set of {project_name} project. Please antgo add product --exp=xxx')
            file_lock.release()
            return False

        if not project_info['auto']:
            logging.warn(f'Project {project_name} not auto optimize. Ignore auto trigger task.')
            file_lock.release()
            return False

        # 发现基于哪个项目实验，进行下一步任务
        product_exp_name = project_info['product']
        exp_info = copy.deepcopy(project_info['exp'][product_exp_name][0])
        if project_info['best']['exp'] != '' and project_info['best']['id'] != '':
            exp_info = copy.deepcopy( project_info['best'])
        exp_info['id'] = ''         # 清空exp id

        if project_event == 'train/label' or project_event == 'product':
            logging.info(f'Trigger {project_event}')
            # 启动监督训练，训练产品模型
            exp_info['checkpoint'] = '' # 清空checkpoint

            next_exp_stage = 'supervised'
            if next_exp_stage not in self.task_set:
                logging.info(f'Add {next_exp_stage} into task queue.')
                self.task_set.add(next_exp_stage)
                self.task_queue.put((self.task_priority[next_exp_stage], next_exp_stage))     

            # 调度
            update_product_info = self.schedule(project_name, project_info, exp_info, auto_find_next_task=False)
            if update_product_info is not None:
                project_info = update_product_info

            # update project info
            with open(os.path.join(config.AntConfig.task_factory, f'{project_name}.json'), 'w') as fp:
                json.dump(project_info, fp)
            file_lock.release()
            return True
        elif project_event == 'train/unlabel':
            logging.info(f'Trigger {project_event}')
            # 启动主动学习，挑选等待标注样本
            next_exp_stage = 'activelearning'
            if len(project_info['tool']['activelearning']['config']) > 0:
                # 存在主动学习的配置方案
                if next_exp_stage not in self.task_set:
                    logging.info(f'Add {next_exp_stage} into task queue.')
                    self.task_set.add(next_exp_stage)
                    self.task_queue.put((self.task_priority[next_exp_stage], next_exp_stage))       

            # 启动半监督学习，直接利用无标签样本训练产品模型
            next_exp_stage = 'semi-supervised'
            if len(project_info['tool']['semi']['config']) > 0:
                # 存在半监督学习的配置方案
                if next_exp_stage not in self.task_set:
                    logging.info(f'Add {next_exp_stage} into task queue.')
                    self.task_set.add(next_exp_stage)
                    self.task_queue.put((self.task_priority[next_exp_stage], next_exp_stage))  

            # 调度
            update_product_info = self.schedule(project_name, project_info, exp_info, auto_find_next_task=False)
            if update_product_info is not None:
                project_info = update_product_info

            # update project info            
            with open(os.path.join(config.AntConfig.task_factory, f'{project_name}.json'), 'w') as fp:
                json.dump(project_info, fp)        

            file_lock.release()
            return True

        file_lock.release()
        return False

    def schedule(self, project_name, project_info, exp_info, auto_find_next_task=True):
        # exp_info 是当前已经运行完成的实验
        # 可以从project_info中获得，当前完成的实验处在哪个阶段
        exp_name = exp_info['exp']
        exp_id = exp_info['id']

        if not project_info['product'].startswith(exp_name):
            logging.warn(f'Exp {exp_name} not project {project_name} product model')
            return None

        if exp_name not in project_info['exp']:
            logging.warn(f'Exp {exp_name} not in project {project_name}.')
            return None

        if auto_find_next_task:
            for check_exp_info in project_info['exp'][exp_name]:
                if check_exp_info['id'] == exp_id:
                    exp_stage = check_exp_info['stage']
                    for order_info in self.task_order:
                        if exp_stage == order_info[0]:
                            next_exp_stage = order_info[1]
                            # 加入任务队列，如果任务队列中存在相同阶段实验，则放弃加入
                            if next_exp_stage not in self.task_set:
                                self.task_set.add(next_exp_stage)
                                self.task_queue.put((self.task_priority[next_exp_stage], next_exp_stage))

        # 根据计算资源进行任务提交
        submitter_info = project_info['submitter']
        if len(submitter_info) == 0:
            # 没有提交配置，清空任务队列
            self.task_set.clear()
            while self.task_queue.qsize():
                self.task_queue.get()
            return

        submitter_method = submitter_info.get('method', '')
        if submitter_method == '':
            # 没有提交配置，清空任务队列
            self.task_set.clear()
            while self.task_queue.qsize():
                self.task_queue.get()
            return        

        submitter_gpu_num = submitter_info['gpu_num']
        submitter_cpu_num = submitter_info['cpu_num']
        submitter_memory = submitter_info['memory']

        # # debug
        # submitter_gpu_num = 0
        # submitter_method = 'local'

        # submitter_resource_check 检查资源
        submitter_resource_check = getattr(script, f'{submitter_method}_submit_resource_check_func', None)
        if submitter_resource_check is None:
            logging.warn(f'Submitter {submitter_method} not support.')
            return None

        # submitter_func 已经决定了如何提交任务
        submitter_func = getattr(script, f'{submitter_method}_submit_process_func', None)
        if submitter_func is None:
            logging.warn(f'Submitter {submitter_method} not support.')
            return None

        # 从任务队列中尝试取出所有等待任务进行提交
        while self.task_queue.qsize() > 0:
            if not submitter_resource_check(submitter_gpu_num, submitter_cpu_num, submitter_memory):
                logging.warn(f'Submitter {submitter_method} resource (gpu: {submitter_gpu_num}, cpu: {submitter_cpu_num}, memory: {submitter_memory}) not enough.')
                return None

            next_task = self.task_queue.get()
            next_task_cmd = self.task_cmd[next_task[1]]
            if next_task[1] == 'label':
                next_task_cmd += f" --exp={exp_name}.{exp_id}"
                next_task_cmd += f" --stage={next_task[1]}"
                next_task_cmd += f" --root={self.root}/{project_name}"
                next_task_cmd += f" --gpu-id=-1"
                # 任务的标注类别信息，标注类型，标注meta信息
                label_tags = ','.join([f'{tag}:{tag_i}' for tag_i, tag in enumerate(project_info['tool']['label']['category'])])
                label_type = project_info['tool']['label']['type']
                next_task_cmd += f" --tags={label_tags}"
                next_task_cmd += f" --type={label_type}"
                if self.ext_module is not None:
                    next_task_cmd += f" --ext-module={self.ext_module}"

                # 创建临时目录
                old_folder = os.path.abspath(os.curdir)
                temp_folder = f'{exp_name}.{exp_id}.label'
                if not os.path.exists(temp_folder):
                    os.makedirs(temp_folder)
                os.chdir(temp_folder)

                logging.info(f'Submit task command {next_task_cmd}')
                if submitter_func(project_name, next_task_cmd, submitter_gpu_num, submitter_cpu_num, submitter_memory, next_task[1]):
                    # 提交任务成功
                    logging.info(f"Success submit label task.")
                    # 修改当前实验的状态，标记为标注（仅对上一阶段为activelearning任务有效）
                    # 其余任务，在调度后不对原有任务的状态进行修正
                    # exp_info['stage'] = 'label'       # 调整当前实验阶段为标注阶段
                    # exp_info['state'] = 'running'     # 调整状态为运行状态
                    for check_exp_info in project_info['exp'][exp_name]:
                        if check_exp_info['id'] == exp_id:
                            check_exp_info['stage'] = 'label'               # 调整当前实验阶段为标注阶段
                            check_exp_info['state'] = 'running'             # 调整状态为运行状态

                    # 将next_task 从任务集合中删除 （任务队列已经没有此任务）
                    self.task_set.remove(next_task[1])
                else:
                    # 提交任务失败
                    logging.error(f'Fail submit label task.')
                    # 将next_task重新加入队列 （任务集合保持不变）
                    self.task_queue.put(next_task)
                    
                # 恢复原状
                # 切换回之前的目录为当前目录（仅对当前线程有效）
                os.chdir(old_folder) 
                # 对于local模型，不进行删除操作      
                if submitter_method != 'local':
                    if temp_folder is not None:
                        shutil.rmtree(temp_folder)
                return project_info

            # 其余涉及模型的任务（训练，预测等）
            next_task_cmd += f" --exp={exp_name}"
            next_task_cmd += f" --stage={next_task[1]}"
            next_task_cmd += f" --root={self.root}/{project_name}"
            # 基于提交脚本设置gpu_ids
            gpu_ids = ','.join([str(i) for i in range(submitter_gpu_num)])
            if submitter_gpu_num == 0:
                gpu_ids = '-1'
            next_task_cmd += f" --gpu-id={gpu_ids}"
            next_task_cmd += f" --checkpoint={exp_info['checkpoint']}"

            # 创建新实验记录
            auto_exp_info = exp_basic_info()
            auto_exp_info['exp'] = exp_info['exp']
            auto_exp_info['id'] = time.strftime('%Y-%m-%dx%H-%M-%S',time.localtime(time.time()))
            auto_exp_info['branch'] = exp_info['branch']
            auto_exp_info['commit'] = exp_info['commit']
            auto_exp_info['state'] = 'running'
            auto_exp_info['stage'] = next_task[1]
            auto_exp_info['checkpoint'] = exp_info['checkpoint']
            next_task_cmd += f" --id={auto_exp_info['id']}"

            logging.info(f'Submit task command {next_task_cmd}')
            
            # 准备代码环境
            old_folder = os.path.abspath(os.curdir)
            git_folder = None
            if os.path.exists(os.path.join(old_folder, '.git')):
                # 当前目录下存在git记录，默认为当前项目实际目录地址
                logging.warn(f'In current default folder {old_folder}, exists some project, wouldnt git clone from {project_info["git"]} again')
            else:
                # git_folder == project_name                
                git_folder = project_info["git"].split('/')[-1].split('.')[0]
                if not os.path.exists(f'./{project_name}'):
                    os.system(f'git clone {project_info["git"]}')    
                # 切换git_folder为当前目录（仅对当前线程有效）
                os.chdir(git_folder)

            if submitter_func(project_name, next_task_cmd, submitter_gpu_num, submitter_cpu_num, submitter_memory, next_task[1]):
                # 提交任务成功
                logging.info(f"Success submit task {self.task_cmd[next_task[1]]} of project {project_name}")

                # 将新增实验加入项目中
                project_info['exp'][exp_info['exp']].append(
                    auto_exp_info
                )

                # 将next_task 从任务集合中删除 （任务队列已经没有此任务）
                self.task_set.remove(next_task[1])
            else:
                # 提交任务失败 (重新加入任务队列)
                logging.error(f'Fail submit task {self.task_cmd[next_task[1]]} of project {project_name}')
                # 将next_task重新加入队列 （任务集合保持不变）
                self.task_queue.put(next_task)

            # 恢复原状
            if git_folder is not None:
                # 切换回之前的目录为当前目录（仅对当前线程有效）
                os.chdir(old_folder) 
            # 对于local模型，不进行删除操作      
            if submitter_method != 'local':
                if git_folder is not None:
                    shutil.rmtree(git_folder)

            return project_info


'''
默认项目全部信息存储在hdfs上
自动化触发, 依靠定时读取hdfs地址实现
'''
class LocalServer(ServerBase):
    def __init__(self, *args, **kwargs) -> None:
        super(LocalServer, self).__init__(*args, **kwargs)
        # antgo 监控root地址，定时从此目录下获得项目更新信息
        self.root = kwargs['root']
        # 目录结构如下
        # root 
        #   - project
        #        - activelearning
        #           - exp.id        # 在此实验之后，触发的主动学习进行标注样本挑选
        #               annotation.json
        #           - exp.id
        #               annotation.json
        #        - label
        #           - exp.id        # 针对主动学习挑选出来的样本，进行标注结果
        #               data.tar    # 压缩包数据格式 images/, annotation.json, meta.json
        #           - exp.id        # 在此实验后，进行的标注结果
        #               data.tar    # 压缩包数据格式 images/, annotation.json, meta.json
        #        - dataset
        #           - label
        #               exp-id-{}-tfrecord    # 针对标注后样本进行的重新打包
        #           - pseudo-label
        #               xxx.tfrecord          # 针对算法生成的伪标签数据进行的打包
        #           - unlabel
        #               xxx.tfrecord          # 针对上传的无标签数据
        #        - exp.id
        #               RUNNING/FINISH/STOP
        #               - output
        #                   - metric
        #                       best.pth
        #                       best.json
        #                   - checkpoint
        #                       ***.pth
        #        - exp.id
        #               RUNNING/FINISH/STOP
        #               - output
        #                   - metric
        #                       best.pth
        #                       best.json
        #                   - checkpoint
        #                       ***.pth        
        #   - project
        #        - exp.id

    def watch(self):
        # 获得所有项目配置信息
        file_lock.acquire()
        for project_file in os.listdir(config.AntConfig.task_factory):
            if not project_file.endswith('.json'):
                continue
            
            with open(os.path.join(config.AntConfig.task_factory, project_file), 'r') as fp:
                project_info = json.load(fp)

            # 项目名称    
            project_name = project_file.replace('.json', '')

            # 检查项目下所有任务的运行情况，对于完成的任务启动下一轮任务调度
            for exp_name, exp_list_info in project_info['exp'].items():
                for exp_info in exp_list_info:
                    exp_id = exp_info['id']
                    if exp_info['stage'] == 'label' and exp_info['state'] == 'running':
                        # 对于在运行的标注任务
                        project_state_folder = os.path.join(self.root, project_name, 'label', f'{exp_name}.{exp_id}')
                        is_exist = file_client_exists(os.path.join(project_state_folder, 'FINISH'))
                        if is_exist:
                            # 标注任务完成，开始调度下一轮任务
                            logging.info(f'Finish label task {exp_name}.{exp_id} project {project_name}.')
                            exp_info['state'] = 'finish'

                            # 获得新打包的标注数据集
                            # root, dataset, label
                            labeled_dataset_records = file_client_ls(os.path.join(self.root, project_name, 'dataset', 'label'))
                            if len(labeled_dataset_records) > 0:
                                existed_dataset_records = [
                                    aabb['address'] for aabb in project_info['dataset']['train']['label']
                                ]
                                temp_set = set()
                                for data_record_file in labeled_dataset_records:
                                    # 同一个数据文件，存在*-index和*-tfrecord
                                    finding_file = '-'.join(data_record_file.split('-')[:-1])
                                    if finding_file in temp_set:
                                        continue

                                    temp_set.add(finding_file)
                                    if finding_file not in existed_dataset_records:
                                        project_info['dataset']['train']['label'].append(
                                            {
                                                'tag': '',
                                                'num': 0,
                                                'status': True,
                                                'address': finding_file
                                            }
                                        )

                            # 保存项目新信息(在任务调度时可能需要重新读取项目信息，提前把更新内容保存)
                            with open(os.path.join(config.AntConfig.task_factory, project_file), 'w') as fp:
                                json.dump(project_info, fp)

                            # 项目自动化优化流水线(仅对产品模型进行自动化优化)
                            if project_info['auto'] and project_info['product'].startswith(exp_name):
                                logging.info(f'Schedule next task of {exp_name}.{exp_id} project {project_name}.')
                                update_product_info = self.schedule(project_name, project_info, exp_info)
                                if update_product_info is not None:
                                    project_info = update_product_info

                    elif exp_info['stage'] != 'label' and exp_info['state'] == 'running':
                        # 对于在运行的非标注任务
                        project_state_folder = os.path.join(self.root, project_name, f'{exp_name}.{exp_id}')
                        checkpoint_folder = os.path.join(self.root, project_name, f'{exp_name}.{exp_id}','output', 'checkpoint')
                        metric_folder = os.path.join(self.root, project_name, f'{exp_name}.{exp_id}','output', 'metric')

                        # FINISH 仅仅表示任务运行完毕
                        # 对于具体完成的任务是什么需要从project_info中查找
                        is_exist = file_client_exists(os.path.join(project_state_folder, 'FINISH'))
                        if is_exist:
                            # 任务完成状态
                            logging.info(f'Finish stage {exp_info["stage"]} task {exp_name}.{exp_id} project {project_name}.')
                            exp_info['state'] = 'finish'
                            exp_info['finish_time'] = time.strftime('%Y-%m-%dx%H-%M-%S',time.localtime(time.time()))

                            # 获得任务保存的checkpoint
                            if file_client_exists(os.path.join(metric_folder, 'best.pth')):
                                # 记录最佳指标模型（开启评估过程）
                                exp_info['checkpoint'] = os.path.join(metric_folder, 'best.pth')
                            elif file_client_exists(os.path.join(checkpoint_folder, 'latest.pth')):
                                # 记录最后的一个模型（未开启评估过程）
                                exp_info['checkpoint'] = os.path.join(checkpoint_folder, 'latest.pth')

                            # 获得任务保存的最佳指标结果
                            best_metric_record = os.path.join(metric_folder, 'best.json')
                            if file_client_exists(best_metric_record):
                                logging.info(f'Finding metric best.json of task {exp_name}.{exp_id}.')
                                if not os.path.exists('./temp'):
                                    os.makedirs('./temp')

                                file_client_get(best_metric_record, './temp')
                                if os.path.exists(os.path.join('./temp/best.json')):
                                    # 读取并更新实验指标信息
                                    with open('./temp/best.json', 'r') as fp:
                                        exp_info['metric'].update(json.load(fp))
                                    
                                    # 清理临时文件
                                    os.remove('./temp/best.json')

                                # 自动更新最佳模型（仅针对产品模型记录）
                                try:
                                    if project_info['product'].startswith(exp_name) and len(exp_info['metric']) > 0:
                                        # 当前完成的实验，存在指标计算，并且当前实验是产品模型
                                        logging.info(f'Update best metric of task {exp_name}.{exp_id} project {project_name}.')
                                        if len(project_info['best']) == 0 or len(project_info['best']['metric']) == 0:
                                            project_info['best'] = exp_info
                                        else:
                                            best_metric_keys = project_info['best']['metric']['score'].keys()
                                            exp_metric_keys = exp_info['metric']['score'].keys()
                                            if set(best_metric_keys) == set(exp_metric_keys):
                                                is_best = False
                                                for k in exp_info['metric']['score'].keys():
                                                    if exp_info['metric']['score'][k] < project_info['best']['metric']['score'][k]:
                                                        is_best = False
                                                
                                                if is_best:
                                                    project_info['best'] = exp_info
                                            else:
                                                logging.error('Abnormal exp metric keys not consistent with best record.')
                                except:
                                    logging.error("Abnormal analyze project best record.")

                            # 获得任务生成的数据（伪标签数据）
                            pseudo_label_dataset_records = file_client_ls(os.path.join(self.root, project_name, 'dataset', 'pseudo-label'))
                            if len(pseudo_label_dataset_records) > 0:
                                existed_dataset_records = [
                                    aabb['address'] for aabb in project_info['dataset']['train']['pseudo-label']
                                ]
                                temp_set = set()
                                for data_record_file in pseudo_label_dataset_records:
                                    # 同一个数据文件，存在*-index和*-tfrecord
                                    finding_file = '-'.join(data_record_file.split('-')[:-1])
                                    if finding_file in temp_set:
                                        continue

                                    temp_set.add(finding_file)
                                    if finding_file not in existed_dataset_records:
                                        project_info['dataset']['train']['label'].append(
                                            {
                                                'tag': '',
                                                'num': 0,
                                                'status': True,
                                                'address': finding_file
                                            }
                                        )

                            # 保存项目新信息(在任务调度时可能需要重新读取项目信息，提前把更新内容保存)
                            with open(os.path.join(config.AntConfig.task_factory, project_file), 'w') as fp:
                                json.dump(project_info, fp)
     
                            # 项目自动化优化流水线(仅对产品模型进行自动化优化)
                            if project_info['auto'] and project_info['product'].startswith(exp_name):
                                logging.info(f'Schedule next task of {exp_name}.{exp_id} project {project_name}.')
                                update_product_info = self.schedule(project_name, project_info, exp_info)
                                if update_product_info is not None:
                                    project_info = update_product_info                                

            # 保存项目新信息
            with open(os.path.join(config.AntConfig.task_factory, project_file), 'w') as fp:
                json.dump(project_info, fp)
        
        file_lock.release()
        # 重新启动下一轮触发
        periodical_func = threading.Timer(self.timer, self.watch)
        periodical_func.start()


class HttpServer(ServerBase):
    def __init__(self) -> None:
        super().__init__()


class ClientHandler(object):
    def __init__(self):
        super().__init__()
        self.host = 'localhost'
        # 默认端口
        self.port = 8886

        # 检查服务保存的端口信息
        config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
        config.AntConfig.parse_xml(config_xml)

        if os.path.exists(os.path.join(config.AntConfig.factory, '.server.info')):   
            with open(os.path.join(config.AntConfig.factory, '.server.info'), 'r') as fp:
                content = json.load(fp)
                self.port = content['port']    

    def alive(self):
        try:
            client = socket(AF_INET, SOCK_STREAM)
            client.connect((self.host, self.port))

            event = 'ping'
            client.send(event.encode('utf-8'))
            response = client.recv(1024)
            if not response:
                logging.error(f'Trigger event {event} fail.')
                return False

            return True        
        except:
            return False

    def trigger(self, event):
        try:
            client = socket(AF_INET, SOCK_STREAM)
            client.connect((self.host, self.port))

            client.send(event.encode('utf-8'))
            response = client.recv(1024)
            if not response:
                logging.error(f'Trigger event {event} fail.')
                return

            response = json.loads(response.decode('utf-8'))
            if response['status']:
                logging.info(f'Success to trigger {event}')
            else:
                logging.error(f'Abnormal to trigger {event}')

            client.close()
        except:
            logging.error(f'Trigger {event} fail (couldnt connect server).')


def get_client():
    client_handler = ClientHandler()
    return client_handler

def get_lock():
    global file_lock
    return file_lock

def launch_server(port, root, ext_module):
    # step1: 尝试使用httpserver，使用MLTALKER服务管理
    # step2: 使用localserver，常用于公网连接受限情况，基于文件服务实现同步
    host = ''
    port = 8886 if port == 0 else port

    # 检查是否已经存在服务进程
    if get_client().alive():
        logging.warn('Antgo server has running in background.')
        return    

    try:
        global server_manager
        server_manager = LocalServer(root=root, ext_module=ext_module)
        server = socketserver.TCPServer((host, port), BackgroundTCP)

        config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
        config.AntConfig.parse_xml(config_xml)

        with open(os.path.join(config.AntConfig.factory, '.server.info'), 'w') as fp:
            fp.write(json.dumps({
                'port': port
            }))
        logging.info(f'Antgo server on host {host} port {port}')
        server.serve_forever()
    except KeyboardInterrupt:
        logging.error('Antgo server CTRL+C,break.')
        sys.exit()      
    except:   
        logging.error('Launch antgo server fail.')
        sys.exit()


if __name__ == '__main__':
    DEFINE_int('port', '8886', '')
    DEFINE_nn_args()
    
    args = parse_args()
    if args.ext_module != '':
        logging.info('import extent module')
        load_extmodule(args.ext_module)

    # 启动服务        
    launch_server(args.port, args.root, args.ext_modeule)


# 优化器配置
optimizer = dict(type='SGD', lr=0.01,  weight_decay=5e-4, momentum=0.01, nesterov=True)
optimizer_config = dict(grad_clip=dict(max_norm=10))

# 学习率调度配置
lr_config = dict(
    policy='CosineAnnealing',
    min_lr=1e-5,
)

# 日志配置
log_config = dict(
    interval=1,    
    hooks=[
        dict(type='TextLoggerHook'),
    ])

# 模型配置
# model 字段根据具体模型进行设置
model = dict(
    type='DenseTeacher',
    train_cfg=dict(
        key=0,
        use_sigmoid=True,
        label_batch_size=128,
        unlabel_batch_size=16,
        semi_ratio=0.5,
        heatmap_n_thr=0.25,
        semi_loss_w=1.0
    ),
    test_cfg=None,
    init_cfg=None,
)

# 自定义 hooks配置
custom_hooks = dict(
    type='MeanTeacher'
)

# 数据配置
data=dict(
    train=[
        dict(
            type="TFDataset",
            data_path_list=[],
            pipeline=[
                dict(type='DecodeImage', to_rgb=True),
                dict(type="KeepRatio", aspect_ratio=1.77),
                dict(type="Rotation", degree=15, border_value=128),            
                dict(type='ResizeS', target_dim=[448,256]),            
                dict(type='ColorDistort'),       
                dict(type='INormalize', mean=[128.0,128.0,128.0], std=[128.0,128.0,128.0],to_rgb=False, keys=['image']),
                dict(type='IImageToTensor', keys=['image']),            
            ],
            inputs_def={
                'fields': ["image", 'bboxes', 'labels', 'image_meta']
            },
            shuffle_queue_size=2048
        ),
        dict(
            type="TFDataset",
            data_path_list=[],
            weak_pipeline=[
                dict(type='DecodeImage', to_rgb=True),
                dict(type="KeepRatio", aspect_ratio=1.77),
                dict(type='ResizeS', target_dim=[448,256]),     
            ],
            strong_pipeline=[
                dict(type='ColorDistort'),       
            ],
            pipeline=[
                dict(type='INormalize', mean=[128.0,128.0,128.0], std=[128.0,128.0,128.0],to_rgb=False, keys=['image']),
                dict(type='IImageToTensor', keys=['image']),            
            ],
            inputs_def={
                'fields': ["image", 'bboxes', 'labels', 'image_meta']
            },
            shuffle_queue_size=2048
        ),            
    ],
    train_dataloader=dict(
        samples_per_gpu=[128,16], 
        workers_per_gpu=4,
        drop_last=True,
        shuffle=True,
        ignore_stack=['bboxes', 'labels', 'image_meta']
    )
)


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/18 13:31
# @File    : demo.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os
import queue
import shutil
import threading
import concurrent.futures
import uuid

from antgo.pipeline.functional.entity import Entity
from antgo.pipeline.functional.option import Some
from .serve import _APIWrapper,_PipeWrapper, _decode_content
from fastapi.staticfiles import StaticFiles
from fastapi.responses import HTMLResponse, FileResponse
from fastapi.templating import Jinja2Templates
from fastapi import Response
from fastapi import File, UploadFile
from fastapi.middleware.cors import CORSMiddleware
import json
import numpy as np
import cv2


class DemoMixin:
  def demo(self, input=[], output=[], app=None, default_config=None):
    if app is None:
      from fastapi import FastAPI, Request
      app = FastAPI()
    else:
      from fastapi import Request
    api = _APIWrapper.tls.place_holder
    pipeline = _PipeWrapper(self._iterable, api)

    input_selection = [cc['data'] for cc in input]
    input_selection_types = [cc['type'] for cc in input]
    for ui_type in input_selection_types:
      assert(ui_type in ['image', 'video', 'text', 'slider', 'checkbox', 'select'])
    
    output_selection = [cc['data'] for cc in output]
    output_selection_types = [cc['type'] for cc in output]
    for ui_type in output_selection_types:
      assert (ui_type in ['image', 'video', 'text', 'number'])

    input_config = default_config
    if default_config is None:
      input_config = [{} for _ in range(len(input_selection))]

    dump_folder = './dump'
    if not os.path.exists(dump_folder):
      os.makedirs(dump_folder)

    resource_dir = '/'.join(os.path.dirname(__file__).split('/')[0:-3])
    static_folder = os.path.join(dump_folder, 'demo', 'static')
    if os.path.exists(static_folder):
      shutil.rmtree(static_folder)
    shutil.copytree(os.path.join(resource_dir, 'resource', 'app'), static_folder)

    if not os.path.exists(os.path.join(static_folder, 'image')):
      os.makedirs(os.path.join(static_folder, 'image'))

    app.add_middleware(
      CORSMiddleware,
      # 允许跨域的源列表，例如 ["http://www.example.org"] 等等，["*"] 表示允许任何源
      allow_origins=["http://localhost:8080"],
      # 跨域请求是否支持 cookie，默认是 False，如果为 True，allow_origins 必须为具体的源，不可以是 ["*"]
      allow_credentials=True,
      # 允许跨域请求的 HTTP 方法列表，默认是 ["GET"]
      allow_methods=["*"],
      # 允许跨域请求的 HTTP 请求头列表，默认是 []，可以使用 ["*"] 表示允许所有的请求头
      # 当然 Accept、Accept-Language、Content-Language 以及 Content-Type 总之被允许的
      allow_headers=["*"],
      # 可以被浏览器访问的响应头, 默认是 []，一般很少指定
      # expose_headers=["*"]
      # 设定浏览器缓存 CORS 响应的最长时间，单位是秒。默认为 600，一般也很少指定
      # max_age=1000
    )

    @app.post('/antgo/api/demo/submit/')
    async def wrapper(req: Request):
      nonlocal pipeline
      req = await _decode_content(req)
      req = json.loads(req['query'])
      for i, b in enumerate(input_selection_types):
        if b in ['image', 'video', 'file']:
          req[i] = os.path.join(dump_folder, req[i])
        if b == 'checkbox':
          req[i] = bool(req[i])

      if len(req) == 1:
        req = req[0]
      rsp = pipeline.execute(req)
      # 输出与类型对齐
      rsp_value = rsp.get()
      output_info = {}
      for i, b in enumerate(output_selection):
        if output_selection_types[i] in ['image', 'video', 'file']:
          value = rsp_value.__dict__[b]
          if type(value) == str:
            shutil.copyfile(value, os.path.join(static_folder, 'image'))
            file_name = value.split('/')[-1]
            value = f'image/{file_name}'
          else:
            if value.dtype == np.uint8:
              transfer_result = value
            else:
              data_min = np.min(value)
              data_max = np.max(value)
              transfer_result = ((value - data_min) / (data_max - data_min) * 255).astype(np.uint8)

            if len(value.shape) == 3:
              assert (value.shape[2] == 3 or value.shape[2] == 4)

            assert (len(value.shape) == 2 or len(value.shape) == 3)
            file_name = f'{uuid.uuid4()}.png'
            cv2.imwrite(os.path.join(static_folder, 'image', file_name), transfer_result)
            value = f'image/{file_name}'

          output_info[b] = {
            'type': output_selection_types[i],
            'name': b,
            'value': value
          }
        else:
          output_info[b] = {
            'type': output_selection_types[i],
            'name': b,
            'value': rsp_value.__dict__[b]
          }
      return output_info

    @app.get('/')
    async def home(request: Request):
      return FileResponse(os.path.join(static_folder, 'index.html'))

    @app.get('/antgo/api/info/')
    async def info():
      return {
        'status': 'OK',
        'message': '',
        'content': {
          'project_type': 'DEMO',
          'project_state': {}
        }
      }

    @app.get('/antgo/api/user/info/')
    async def user_info():
      return {
        'status': 'OK',
        'message': '',
        'content': {
          'user_name': 'ANTGO',
          'short_name': 'A',
          'task_name': 'DEFAULT',
          'task_type': 'DEFAULT',
          'project_type': 'DEMO',
        }
      }

    @app.post('/antgo/api/demo/upload/')
    async def upload(file: UploadFile = File(...)):
      try:
        contents = file.file.read()
        with open(os.path.join(dump_folder, file.filename), 'wb') as f:
          f.write(contents)
      except Exception:
        return {"message": "There was an error uploading the file"}
      finally:
        file.file.close()

    @app.get('/antgo/api/demo/query_config/')
    async def query_config():
      input_info = []
      for k, v, config in zip(input_selection, input_selection_types, input_config):
        info = {
          'type': v,
          'name': k,
          'value': ''
        }
        if v == 'text':
          info['value'] = config.get('value', '')
        if v == 'slider':
          info['value'] = config.get('value', 0)
          info['min'] = config.get('min', 0)
          info['max'] = config.get('max', 100)
        if v == 'checkbox':
          info['value'] = (int)(config.get('value', False))
        if v == 'select':
          info['value'] = config.get('value', '')
          info['options'] = config.get('options', [])
          assert(len(info['options']) >= 1)
          value_list = [option['value'] for option in info['options']]
          assert(info['value'] in value_list)

        input_info.append(info)

      return input_info

    # static resource
    app.mount("/", StaticFiles(directory=static_folder), name="static")
    return app

  @classmethod
  def web(cls, index=None):
    return _APIWrapper(index=index, cls=cls)


import os
import requests
from antgo.ant import flags
import time
import  urllib.parse
import uuid
from antgo.utils import logger

def get_file_name(headers):
    filename = ''
    if 'Content-Disposition' in headers and headers['Content-Disposition']:
        disposition_split = headers['Content-Disposition'].split(';')
        if len(disposition_split) > 1:
            if disposition_split[1].strip().lower().startswith('filename='):
                file_name = disposition_split[1].split('=')
                if len(file_name) > 1:
                    filename = urllib.parse.unquote(file_name[1])
    if filename == '':
        filename = str(uuid.uuid4())
    
    return filename


def activelearning_api_download():
    FLAGS = flags.AntFLAGS

    # 服务ip:port
    host_ip = FLAGS.host_ip()
    host_port = FLAGS.host_port()
    round = None
    for p in FLAGS.param().split(','):
        k,v = p.split(":")
        if k == 'round':
            round = (int)(v)

    if round is None:
        logger.error('Need set param (eg. --param=round:0)')
        return    

    download_url = 'http://%s:%d/activelearning/download/'%(host_ip, host_port)
    down_res = requests.get(url=download_url, params={'round': round})
    file_name = get_file_name(down_res.headers)

    with open(file_name, "wb") as code:
        code.write(down_res.content)


def activelearning_api_upload():
    FLAGS = flags.AntFLAGS

    # 上传的文件路径
    file_path = FLAGS.file()

    # 服务ip:port
    host_ip = FLAGS.host_ip()
    host_port = FLAGS.host_port()

    round = None
    for p in FLAGS.param().split(';'):
        k,v = p.split(":")
        if k == 'round':
            round = (int)(v)
    
    if round is None:
        logger.error('Need set param (eg. --param=round:0)')
        return    

    if not os.path.exists(file_path):
        logger.error('%s file dont exist.'%file_path)
        return

    file_name = os.path.normpath(file_path).split('/')[-1]
    file_size = os.path.getsize(file_path)
    chunk_size = 1024*1024 # 1M
    chunk_num = (file_size + chunk_size - 1)//chunk_size
    finished_size = 0
    api_url = 'http://%s:%d/activelearning/upload/'%(host_ip, host_port)
    with open(file_path, 'rb') as fp:
        for chunk_id in range(chunk_num):
            remained_size = chunk_size
            if finished_size + chunk_size < file_size:
                remained_size = chunk_size
            else:
                remained_size = finished_size+chunk_size - file_size
            content = fp.read(remained_size)
            kwargs = {}
            kwargs.update({'sliceIndex': chunk_id, 'sliceNum': chunk_num, 'fileName': file_name, 'round': round})

            result = requests.post(api_url, kwargs, files={'file': content})
            if result.status_code not in [200, 201]:
                logger.error('Upload fail, exit.')
                return


# Copyright (c) OpenMMLab. All rights reserved.
from ..builder import BBOX_CODERS
from .base_bbox_coder import BaseBBoxCoder


@BBOX_CODERS.register_module()
class PseudoBBoxCoder(BaseBBoxCoder):
    """Pseudo bounding box coder."""

    def __init__(self, **kwargs):
        super(BaseBBoxCoder, self).__init__(**kwargs)

    def encode(self, bboxes, gt_bboxes):
        """torch.Tensor: return the given ``bboxes``"""
        return gt_bboxes

    def decode(self, bboxes, pred_bboxes):
        """torch.Tensor: return the given ``pred_bboxes``"""
        return pred_bboxes



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 17:49
# @File    : __init__.py.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from .base import Operator, NNOperator, PyOperator, OperatorFlag, SharedType

__all__ = [
    'Operator', 'NNOperator', 'PyOperator', 'OperatorFlag',
    'SharedType'
]